#!/usr/bin/env python
"""
_SeederManager_

Util to instantiate a set of feeders based on settings from a WMStep
and then apply them to job groups

"""

from builtins import map, object
from future.utils import viewvalues

from WMCore.JobSplitting.Generators.GeneratorFactory import GeneratorFactory

from WMCore.WMSpec.WMTask import WMTask, WMTaskHelper
from WMCore.WMSpec.ConfigSectionTree import *

class GeneratorManager(object):
    """
    _GeneratorManager_



    """
    def __init__(self, task = None):
        self.generators = {}
        self.generatorFactory = GeneratorFactory()

        if not hasattr(task, 'data'):
            #We don't have a WMTask
            return
        if not hasattr(task.data, 'generators'):
            #We have a blank task with no generators
            return
        #Otherwise we have a fully formed task of some type

        configList = task.listGenerators()

        for generator in configList:
            self.addGenerator(generator, **task.getGeneratorSettings(generator))


        return

    def addGenerator(self, generatorName, **args):
        """
        _addGenerator_

        Add a new instance of the generator provided

        """
        if generatorName == None:
            return
        if generatorName in self.generators:
            return

        #TODO: Exception check
        newGenerator = self.generatorFactory.getGenerator(generatorName, **args)

        self.generators[generatorName] = newGenerator
        return



    def __call__(self, jobGroup):
        """
        _operator(jobGroup)_

        Run all generators in this manager over the jobs in the
        job group provided

        """
        [list(map(generator, jobGroup.jobs)) for generator in viewvalues(self.generators)]
        return


    def getGeneratorList(self):
        """
        _getGeneratorList_

        Returns a list of all generators for usage in JobSplitting
        """
        generatorList = []

        for generator in viewvalues(self.generators):
            generatorList.append(generator)

        return generatorList



#!/usr/bin/env python
"""
_SetLocation_

Oracle implementation of Jobs.SetLocation
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.SetLocation import SetLocation as MySQLSetLocation

class SetLocation(MySQLSetLocation):
    pass



#!/usr/bin/env python
"""
_ReReco_

Standard ReReco workflow.
"""
from __future__ import division

import json

from future.utils import viewitems

from Utils.Utilities import makeList
from WMCore.WMSpec.StdSpecs.DataProcessing import DataProcessing
from WMCore.WMSpec.WMWorkloadTools import validateArgumentsCreate


class ReRecoWorkloadFactory(DataProcessing):
    """
    _ReRecoWorkloadFactory_

    Stamp out ReReco workflows.
    """

    def buildWorkload(self):
        """
        _buildWorkload_

        Build the workload given all of the input parameters.  At the very least
        this will create a processing task and merge tasks for all the outputs
        of the processing task.
        Note that there will be LogCollect tasks created for each processing
        task and Cleanup tasks created for each merge task.
        """
        (self.inputPrimaryDataset, self.inputProcessedDataset,
         self.inputDataTier) = self.inputDataset[1:].split("/")

        workload = self.createWorkload()
        workload.setDashboardActivity("reprocessing")
        workload.setWorkQueueSplitPolicy("Block", self.procJobSplitAlgo,
                                         self.procJobSplitArgs,
                                         OpenRunningTimeout=self.openRunningTimeout)
        procTask = workload.newTask("DataProcessing")

        cmsswStepType = "CMSSW"
        taskType = "Processing"

        forceUnmerged = False
        if self.transientModules:
            # If we have at least one output module not being merged,
            # we must force all the processing task to be unmerged
            forceUnmerged = True

        outputMods = self.setupProcessingTask(procTask, taskType,
                                              self.inputDataset,
                                              couchDBName=self.couchDBName,
                                              configCacheUrl=self.configCacheUrl,
                                              forceUnmerged=forceUnmerged,
                                              configDoc=self.configCacheID,
                                              splitAlgo=self.procJobSplitAlgo,
                                              splitArgs=self.procJobSplitArgs,
                                              stepType=cmsswStepType)
        self.addLogCollectTask(procTask)

        # no real need to sort it, but we better have the same order between Py2/Py3
        for outputModuleName in sorted(list(outputMods)):
            # Only merge the desired outputs
            if outputModuleName not in self.transientModules:
                self.addMergeTask(procTask, self.procJobSplitAlgo, outputModuleName)
            else:
                self.addCleanupTask(procTask, outputModuleName)

        self.addSkims(workload)
        # setting the parameters which need to be set for all the tasks
        # sets acquisitionEra, processingVersion, processingString
        workload.setTaskPropertiesFromWorkload()

        # set the LFN bases (normally done by request manager)
        # also pass runNumber (workload evaluates it)
        workload.setLFNBase(self.mergedLFNBase, self.unmergedLFNBase,
                            runNumber=self.runNumber)

        return workload

    def addSkims(self, workload):
        """
        _addSkims_

        Add skims to the standard dataprocessing workload that was given.
        Note that there will be LogCollect tasks created for each processing
        task and Cleanup tasks created for each merge task.
        """
        skimmableTasks = {}
        procTask = workload.getTopLevelTask()[0]
        for skimmableTask in procTask.childTaskIterator():
            if skimmableTask.taskType() == "Merge":
                skimmableTasks[skimmableTask.data.input.outputModule] = skimmableTask
        # Now add the output modules that are not merged but may be skimmed
        for outputModule in self.transientModules:
            skimmableTasks[outputModule] = procTask

        for skimConfig in self.skimConfigs:
            skimmableTask = skimmableTasks[skimConfig["SkimInput"]]
            skimTask = skimmableTask.addTask(skimConfig["SkimName"])
            parentCmsswStep = skimmableTask.getStep("cmsRun1")

            skimSizePerEvent = skimConfig["SizePerEvent"]
            skimTimePerEvent = skimConfig["TimePerEvent"]
            skimMemory = skimConfig["Memory"]

            # Check that the splitting agrees, if the parent is event based then we must do WMBSMergeBySize
            # With reasonable defaults
            skimJobSplitAlgo = skimConfig["SkimJobSplitAlgo"]
            skimJobSplitArgs = skimConfig["SkimJobSplitArgs"]
            if skimmableTask.jobSplittingAlgorithm == "EventBased":
                skimJobSplitAlgo = "WMBSMergeBySize"
                skimJobSplitArgs = {"max_merge_size": self.maxMergeSize,
                                    "min_merge_size": self.minMergeSize,
                                    "max_merge_events": self.maxMergeEvents,
                                    "max_wait_time": self.maxWaitTime}

            # Define the input module
            inputModule = "Merged"
            if skimConfig["SkimInput"] in self.transientModules:
                inputModule = skimConfig["SkimInput"]

            outputMods = self.setupProcessingTask(skimTask, "Skim",
                                                  inputStep=parentCmsswStep,
                                                  inputModule=inputModule,
                                                  couchDBName=self.couchDBName,
                                                  configCacheUrl=self.configCacheUrl,
                                                  configDoc=skimConfig["ConfigCacheID"],
                                                  splitAlgo=skimJobSplitAlgo,
                                                  splitArgs=skimJobSplitArgs,
                                                  timePerEvent=skimTimePerEvent,
                                                  sizePerEvent=skimSizePerEvent,
                                                  memoryReq=skimMemory)

            self.addLogCollectTask(skimTask, taskName="%sLogCollect" % skimConfig["SkimName"])

            for outputModuleName in outputMods:
                self.addMergeTask(skimTask, skimJobSplitAlgo, outputModuleName)

        return

    def __call__(self, workloadName, arguments):
        """
        _call_

        Create a ReReco workload with the given parameters.
        """
        DataProcessing.__call__(self, workloadName, arguments)

        # Arrange the skims in a skimConfig object (i.e. a list of skim configurations)
        self.skimConfigs = []
        skimIndex = 1
        while "SkimName%s" % skimIndex in arguments:
            skimConfig = {}
            skimConfig["SkimName"] = arguments["SkimName%s" % skimIndex]
            skimConfig["SkimInput"] = arguments["SkimInput%s" % skimIndex]
            skimConfig["ConfigCacheID"] = arguments["Skim%sConfigCacheID" % skimIndex]
            skimConfig["TimePerEvent"] = float(arguments.get("SkimTimePerEvent%s" % skimIndex, self.timePerEvent))
            skimConfig["SizePerEvent"] = float(arguments.get("SkimSizePerEvent%s" % skimIndex, self.sizePerEvent))
            skimConfig["Memory"] = float(arguments.get("SkimMemory%s" % skimIndex, self.memory))
            skimConfig["SkimJobSplitAlgo"] = arguments.get("SkimSplittingAlgo%s" % skimIndex, "FileBased")
            skimConfig["SkimJobSplitArgs"] = {"include_parents": True}
            if skimConfig["SkimJobSplitAlgo"] == "FileBased":
                skimConfig["SkimJobSplitArgs"]["files_per_job"] = int(arguments.get("SkimFilesPerJob%s" % skimIndex, 1))
            elif skimConfig["SkimJobSplitAlgo"] in ["EventBased", "EventAwareLumiBased"]:
                standardSkim = int((8.0 * 3600.0) / skimConfig["TimePerEvent"])
                skimConfig["SkimJobSplitArgs"]["events_per_job"] = int(arguments.get("SkimEventsPerJob%s" % skimIndex, standardSkim))
                if skimConfig["SkimJobSplitAlgo"] == "EventAwareLumiBased":
                    skimConfig["SkimJobSplitAlgo"]["job_time_limit"] = 48 * 3600  # 2 days
            elif skimConfig["SkimJobSplitAlgo"] == "LumiBased":
                skimConfig["SkimJobSplitArgs"]["lumis_per_job"] = int(arguments.get("SkimLumisPerJob%s" % skimIndex, 8))
            self.skimConfigs.append(skimConfig)
            skimIndex += 1

        return self.buildWorkload()

    @staticmethod
    def getWorkloadCreateArgs():

        baseArgs = DataProcessing.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "ReReco", "optional": False},
                    "TransientOutputModules": {"default": [], "type": makeList,
                                               "attr": "transientModules", "null": False}
                    }
        baseArgs.update(specArgs)
        DataProcessing.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    @staticmethod
    def getSkimArguments():
        """
        _getSkimArguments_

        Skim arguments can be many of the same, it depends on the number
        of defined skims. However we need to keep the same definition of its arguments
        in a generic form. This method follows the same definition of getWorkloadCreateArgs in StdBase.
        """
        skimArgs = {
            "SkimName#N": {"default": None, "type": str,
                           "optional": False, "validate": None,
                           "null": False},
            "SkimInput#N": {"default": None, "type": str,
                            "optional": False, "validate": None,
                            "null": False},
            "Skim#NConfigCacheID": {"default": None, "type": str,
                                    "optional": False, "validate": None,
                                    "null": False},
            "SkimTimePerEvent#N": {"default": None, "type": float,
                                   "optional": True, "validate": lambda x: x > 0,
                                   "null": False},
            "SkimSizePerEvent#N": {"default": None, "type": float,
                                   "optional": True, "validate": lambda x: x > 0,
                                   "null": False},
            "SkimMemory#N": {"default": None, "type": float,
                             "optional": True, "validate": lambda x: x > 0,
                             "null": False},
            "SkimSplittingAlgo#N": {"default": None, "type": str,
                                    "optional": True, "validate": None,
                                    "null": False},
            "SkimEventsPerJob#N": {"default": None, "type": int,
                                   "optional": True, "validate": lambda x: x > 0,
                                   "null": False},
            "SkimLumisPerJob#N": {"default": 8, "type": int,
                                  "optional": True, "validate": lambda x: x > 0,
                                  "null": False},
            "SkimFilesPerJob#N": {"default": 1, "type": int,
                                  "optional": True, "validate": lambda x: x > 0,
                                  "null": False}}
        return skimArgs

    def validateSchema(self, schema):
        """
        _validateSchema_

        Check for required fields, and some skim facts
        """
        DataProcessing.validateSchema(self, schema)
        mainOutputModules = list(self.validateConfigCacheExists(configID=schema["ConfigCacheID"],
                                                           configCacheUrl=schema['ConfigCacheUrl'],
                                                           couchDBName=schema["CouchDBName"],
                                                           getOutputModules=True))

        # Skim facts have to be validated outside the usual master validation
        skimSchema = {k: v for (k, v) in viewitems(schema) if k.startswith("Skim")}
        skimArguments = self.getSkimArguments()
        skimIndex = 1
        skimInputs = set()
        while "SkimName%s" % skimIndex in schema:
            instanceArguments = {}
            for argument in skimArguments:
                realArg = argument.replace("#N", str(skimIndex))
                instanceArguments[realArg] = skimArguments[argument]
            try:
                validateArgumentsCreate(skimSchema, instanceArguments)
                # Validate GPU-related spec parameters
                DataProcessing.validateGPUSettings(schema)
            except Exception as ex:
                self.raiseValidationException(str(ex))

            self.validateConfigCacheExists(configID=schema["Skim%sConfigCacheID" % skimIndex],
                                           configCacheUrl=schema['ConfigCacheUrl'],
                                           couchDBName=schema["CouchDBName"],
                                           getOutputModules=False)
            if schema["SkimInput%s" % skimIndex] not in mainOutputModules:
                error = "Processing config does not have the following output module: %s." % schema[
                    "SkimInput%s" % skimIndex]
                self.raiseValidationException(msg=error)
            skimInputs.add(schema["SkimInput%s" % skimIndex])
            skimIndex += 1

        # Validate that the transient output modules are used in a skim task
        if "TransientOutputModules" in schema:
            diffSet = set(schema["TransientOutputModules"]) - skimInputs
            if diffSet:
                self.raiseValidationException(
                    msg="A transient output module was specified but no skim was defined for it")



#! /bin/env python

"""
A service class for retrieving data from McM using
an SSO cookie since it sits behind CERN SSO
"""

from builtins import str, object

import json
import os
import pycurl
import subprocess

from io import BytesIO

from WMCore.WMException import WMException

class McMNoDataError(WMException):
    """
    _McMNoDataError_
        McM responded but has no data for the request
    """

    def __init__(self):
        WMException.__init__(self, 'McM responded correctly but has no data')

class McM(object):
    """
    A service class for retrieving data from McM using
    an SSO cookie since it sits behind CERN SSO
    'key' must be unencrypted
    """

    def __init__(self, cert, key, url='https://cms-pdmv.cern.ch/mcm', tmpDir='/tmp'):
        self.url = url
        self.tmpDir = tmpDir
        self.cert = cert
        self.key = key
        self.cookieFile = None

    def __enter__(self):
        self.cookieFile = os.path.join(self.tmpDir, 'ssoCookie.txt')
        process = subprocess.Popen(["cern-get-sso-cookie", "--cert", self.cert, "--key", self.key,
                                    "-u", self.url, "-o", self.cookieFile],
                                   stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=False)
        strout = process.communicate()[0]
        if process.returncode != 0:
            raise RuntimeError(" FATAL -- could not generate SSO cookie\nError msg: %s" % (str(strout)))
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if self.cookieFile:
            try:
                os.remove(self.cookieFile)
            except OSError:
                return
        return

    def _getURL(self, extendURL):
        """
        Fetch an MCM URL with CURL using the SSO cookie.
        Only intended to be used internally
        """

        try:
            b = BytesIO()
            c = pycurl.Curl()

            fullUrl = '%s/%s' % (self.url, extendURL)

            c.setopt(c.URL, fullUrl)
            c.setopt(c.SSL_VERIFYPEER, False)
            c.setopt(c.SSL_VERIFYHOST, False)
            c.setopt(c.FOLLOWLOCATION, True)
            c.setopt(c.COOKIEJAR, self.cookieFile)
            c.setopt(c.COOKIEFILE, self.cookieFile)
            c.setopt(c.WRITEFUNCTION, b.write)
            c.perform()
            if c.getinfo(pycurl.HTTP_CODE) != 200:
                raise IOError
        except:
            c.close()
            raise IOError('Was not able to fetch or decode URL from McM')

        try:
            body = b.getvalue()
            res = json.loads(body)
        except ValueError:
            c.close()
            raise IOError('Was not able to decode JSON from McM')

        c.close()
        return res

    def getHistory(self, prepID):
        """
        Get the history record which has who did what to an McM request
        """

        try:
            url = 'search?db_name=batches&contains=%s&get_raw' % prepID
            res = self._getURL(url)
            history = res['rows'][0]['doc']['history']
            return history
        except IndexError:
            raise McMNoDataError

    def getRequest(self, prepID):
        """
        Get the request record which has, among other things,
        the number of requested events
        """

        url = 'public/restapi/requests/get/%s' % prepID
        res = self._getURL(url)
        return res['results']

if __name__ == '__main__':
    with McM(cert='.globus/usercert.pem', key='.globus/nopasskey.pem') as mcm:
        history = mcm.getHistory(prepID = 'BTV-Upg2023SHCAL14DR-00002')
        request = mcm.getRequest(prepID = 'BTV-Upg2023SHCAL14DR-00002')



#!/usr/bin/env python
"""
_ReportEmu_

Class for creating bogus framework job reports.
"""
from __future__ import division

from builtins import object
import os.path

from WMCore.DataStructs.File import File
from WMCore.FwkJobReport import Report
from WMCore.Services.UUIDLib import makeUUID


class ReportEmu(object):
    """
    _ReportEmu_

    Job Report Emulator that creates a Report given a WMTask/WMStep and a Job instance.
    """

    def __init__(self, **options):
        """
        ___init___

        Options contain the settings for producing the report instance from the provided step
        """
        self.step = options.get("WMStep", None)
        self.job = options.get("Job", None)
        return

    def addInputFilesToReport(self, report):
        """
        _addInputFilesToReport_

        Pull all of the input files out of the job and add them to the report.
        """
        report.addInputSource("PoolSource")

        for inputFile in self.job["input_files"]:
            inputFileSection = report.addInputFile("PoolSource", lfn=inputFile["lfn"],
                                                   size=inputFile["size"],
                                                   events=inputFile["events"])
            Report.addRunInfoToFile(inputFileSection, inputFile["runs"])

        return

    def determineOutputSize(self):
        """
        _determineOutputSize_

        Determine the total size of and number of events in the input files and
        use the job mask to scale that to something that would reasonably
        approximate the size of and number of events in the output.
        """
        totalSize = 0
        totalEvents = 0

        for inputFile in self.job["input_files"]:
            totalSize += inputFile["size"]
            totalEvents += inputFile["events"]

        if self.job["mask"]["FirstEvent"] is not None and \
                        self.job["mask"]["LastEvent"] is not None:
            outputTotalEvents = self.job["mask"]["LastEvent"] - self.job["mask"]["FirstEvent"] + 1
        else:
            outputTotalEvents = totalEvents

        outputSize = int(totalSize * outputTotalEvents / totalEvents )
        return (outputSize, outputTotalEvents)

    def addOutputFilesToReport(self, report):
        """
        _addOutputFilesToReport_

        Add output files to every output module in the step.  Scale the size
        and number of events in the output files appropriately.
        """
        (outputSize, outputEvents) = self.determineOutputSize()

        if not os.path.exists('ReportEmuTestFile.txt'):
            with open('ReportEmuTestFile.txt', 'w') as f:
                f.write('A Shubbery')

        for outputModuleName in self.step.listOutputModules():
            outputModuleSection = self.step.getOutputModule(outputModuleName)
            outputModuleSection.fixedLFN = False
            outputModuleSection.disableGUID = False

            outputLFN = "%s/%s.root" % (outputModuleSection.lfnBase,
                                        str(makeUUID()))
            outputFile = File(lfn=outputLFN, size=outputSize, events=outputEvents,
                              merged=False)
            outputFile.setLocation(self.job["location"])
            outputFile['pfn'] = "ReportEmuTestFile.txt"
            outputFile['guid'] = "ThisIsGUID"
            outputFile["checksums"] = {"adler32": "1234", "cksum": "5678"}
            outputFile["dataset"] = {"primaryDataset": outputModuleSection.primaryDataset,
                                     "processedDataset": outputModuleSection.processedDataset,
                                     "dataTier": outputModuleSection.dataTier,
                                     "applicationName": "cmsRun",
                                     "applicationVersion": self.step.getCMSSWVersion()}
            outputFile["module_label"] = outputModuleName

            outputFileSection = report.addOutputFile(outputModuleName, outputFile)
            for inputFile in self.job["input_files"]:
                Report.addRunInfoToFile(outputFileSection, inputFile["runs"])

        return

    def __call__(self):
        report = Report.Report(self.step.name())

        report.id = self.job["id"]
        report.task = self.job["task"]
        report.workload = None

        self.addInputFilesToReport(report)
        self.addOutputFilesToReport(report)
        return report



#!/usr/bin/env python
"""
_LoadFromID_

Oracle implementation of LoadFileset
"""

__all__ = []



from WMCore.WMBS.MySQL.Fileset.LoadFromID import LoadFromID as LoadFilesetMySQL

class LoadFromID(LoadFilesetMySQL):
    sql = LoadFilesetMySQL.sql



#!/usr/bin/env python
"""
_LoadFromID_

MySQL implementation of Jobs.LoadFromID.
"""

from WMCore.Database.DBFormatter import DBFormatter


class LoadFromID(DBFormatter):
    """
    _LoadFromID_

    Retrieve meta data for a job given it's ID.  This includes the name,
    job group and last update time.
    """
    sql = """SELECT wmbs_job.id, jobgroup, wmbs_job.name AS name,
                    wmbs_job_state.name AS state, wmbs_job.state_time, retry_count,
                    couch_record,  cache_dir, wmbs_location.site_name AS location,
                    outcome AS bool_outcome, fwjr_path AS fwjr_path
             FROM wmbs_job
               LEFT OUTER JOIN wmbs_location ON
                 wmbs_job.location = wmbs_location.id
               LEFT OUTER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
             WHERE wmbs_job.id = :jobid"""

    def formatDict(self, result):
        """
        _formatDict_

        Cast the id, jobgroup and last_update columns to integers because
        formatDict() turns everything into strings.
        """

        formattedResult = DBFormatter.formatDict(self, result)

        for entry in formattedResult:
            if entry["bool_outcome"] == 0:
                entry["outcome"] = "failure"
            else:
                entry["outcome"] = "success"

            del entry["bool_outcome"]

        if len(formattedResult) == 1:
            return formattedResult[0]
        else:
            return formattedResult

    def execute(self, jobID, conn=None, transaction=False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """

        if isinstance(jobID, list):
            binds = jobID
        else:
            binds = {"jobid": jobID}

        result = self.dbi.processData(self.sql, binds, conn=conn,
                                      transaction=transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_TestImpl_

Test harness for invoking an Implementation plugin

"""
from __future__ import print_function

import getopt
import sys

from WMCore.Storage.Registry import retrieveStageOutImpl
from WMCore.Storage.StageOutError import StageOutError

valid = ['input-pfn=', 'protocol=', "impl=", "target-pfn=", "cleanup"]
try:
    opts, args = getopt.getopt(sys.argv[1:], "", valid)
except getopt.GetoptError as ex:
    print(str(ex))
    sys.exit(1)

inputPfn = None
targetPfn = None
implName = None
protocol = None
doCleanup = False

for opt, arg in opts:
    if opt == "--input-pfn":
        inputPfn = arg
    if opt == "--target-pfn":
        targetPfn = arg
    if opt == "--impl":
        implName = arg
    if opt == "--protocol":
        protocol = arg
    if opt == "--cleanup":
        doCleanup = True

if implName == None:
    msg = "Error: ImplName not provided, you need to provide the --impl option"
    print(msg)
    sys.exit(1)

if inputPfn == None:
    msg = "Error: Input PFN not provided: use the --input-pfn option"
    print(msg)
    sys.exit(1)

if targetPfn == None:
    msg = "Error: Target PFN not provided: use the --target-pfn option"
    print(msg)
    sys.exit(1)

try:
    implInstance = retrieveStageOutImpl(implName)
except StageOutError as ex:
    print("Error retrieving plugin from registry named: %s" % implName)
    print(str(ex))
    sys.exit(1)

print(" Invoking StageOutImpl: %s" % implName)
implInstance(protocol, inputPfn, targetPfn)

if doCleanup:
    #  //
    # // Cleanup implies that we invoke the implementations
    # //  removeFile method on the target PFN after the transfer
    print("Invoking %s.removeFile on %s" % (implName, targetPfn))
    implInstance.removeFile(targetPfn)



#!/usr/bin/env python
"""
_LoadRunning_

Oracle implementation for loading bl_runjob records for active jobs
"""


from WMCore.BossAir.MySQL.LoadRunning import LoadRunning as MySQLLoadRunning

class LoadRunning(MySQLLoadRunning):
    """
    _LoadRunning_

    """



"""
_GetAllHeartbeatInfo_

Oracle implementation of GetAllHeartbeatInfo
"""

__all__ = []



from WMCore.Agent.Database.MySQL.GetAllHeartbeatInfo import GetAllHeartbeatInfo \
     as GetAllHeartbeatInfoMySQL

class GetAllHeartbeatInfo(GetAllHeartbeatInfoMySQL):
    pass



#!/usr/bin/env python
# encoding: utf-8
"""
CouchFileset.py

Created by Dave Evans on 2010-03-19.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from future.utils import viewvalues

import time

import WMCore.Database.CMSCouch as CMSCouch
from WMCore.ACDC.Fileset import Fileset
from WMCore.Algorithms.ParseXMLFile import coroutine
from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset as DataStructsFileset
from WMCore.DataStructs.Run import Run
from WMCore.Database.CouchUtils import connectToCouch, requireFilesetName


@coroutine
def makeRun(targets):
    """
    _makeRun_

    create a DataStruct Run instance form the Couch JSON dict
    """
    while True:
        fileRef, runs = (yield)
        for run in runs:
            newRun = Run(run[u'Run'], *run[u'Lumis'])
            fileRef.addRun(newRun)


@coroutine
def filePipeline(targets):
    """
    _conversionPipeline_

    """
    while True:
        inputDict = (yield)
        newFile = File(
            lfn=str(inputDict[u'lfn']),
            size=int(inputDict[u'size']),
            events=int(inputDict[u'events'])
        )
        targets['run'].send((newFile, inputDict[u'runs']))
        targets['fileset'].addFile(newFile)


class CouchFileset(Fileset):
    def __init__(self, **options):
        Fileset.__init__(self, **options)
        self.url = options.get('url')
        self.database = options.get('database')
        self['name'] = options.get('name')
        self.server = None
        self.couchdb = None

    @connectToCouch
    def drop(self):
        """
        _drop_

        Remove this fileset

        This is racy. someone can add to the fileset before we get done
        deleting it. Actually, they can delete after we get done deleting
        it too. Oh well.
        """
        for d in self.filelistDocuments():
            self.couchdb.delete_doc(d)

    @connectToCouch
    @requireFilesetName
    def filelistDocuments(self):
        """
        _filelistDocuments_

        Get a list of document ids corresponding to filelists in this fileset
        """
        params = {"startkey": [self.collectionName, self["name"]],
                  "endkey": [self.collectionName, self["name"]],
                  "reduce": False}
        result = self.couchdb.loadView("ACDC", "coll_fileset_docs",
                                       params)

        docs = [row["id"] for row in result["rows"]]
        return docs

    @connectToCouch
    @requireFilesetName
    def add(self, files, mask=None):
        """
        _add_

        Add files to this fileset

        Note: if job was lumi based splitted, then we do not have
        reliable events information. If job was event based splitted,
        then we do not have reliable lumi information.
        """
        filteredFiles = []
        if mask:
            for f in files:
                # There might be no LastEvent for last job of a file
                if mask['LastEvent'] and mask['FirstEvent']:
                    f['events'] = mask['LastEvent'] - mask['FirstEvent'] + 1
                    f['first_event'] = mask['FirstEvent']
                elif mask['FirstEvent']:
                    f['events'] = f['events'] - mask['FirstEvent'] + 1
                    f['first_event'] = mask['FirstEvent']

            maskLumis = mask.getRunAndLumis()
            if maskLumis != {}:
                # Then we actually have to do something
                for f in files:
                    newRuns = mask.filterRunLumisByMask(runs=f['runs'])
                    if newRuns != set([]):
                        f['runs'] = newRuns
                        filteredFiles.append(f)
            else:
                # Likely real data with EventBased splitting
                filteredFiles = files
        else:
            filteredFiles = files

        jsonFiles = {}
        for f in filteredFiles:
            jsonFiles.__setitem__(f['lfn'], f.__to_json__(None))
        fileList = self.makeFilelist(jsonFiles)
        return fileList

    @connectToCouch
    @requireFilesetName
    def makeFilelist(self, files=None):
        """
        _makeFilelist_

        Create a new filelist document containing the id
        """
        files = files or {}
        # add a version to each of these ACDC docs such that we can properly
        # parse them and avoid issues between ACDC docs and agent base code
        input = {"collection_name": self.collectionName,
                 "collection_type": self.collectionType,
                 "fileset_name": self["name"],
                 "files": files,
                 "acdc_version": 2,
                 "timestamp": time.time()}

        document = CMSCouch.Document(None, input)

        commitInfo = self.couchdb.commitOne(document)
        document['_id'] = commitInfo[0]['id']
        if 'rev' in commitInfo[0]:
            document['_rev'] = commitInfo[0]['rev']
        else:
            if commitInfo[0]['reason'].find('{exit_status,0}') != -1:
                # TODO: in this case actually insert succeeded but return error
                # due to the bug
                # https://issues.apache.org/jira/browse/COUCHDB-893
                # if rev is needed to proceed need to get by
                # self.couchdb.documentExist(document['_id'])
                # but that function need to be changed to return _rev
                document['_rev'] = "NeedToGet"
            else:
                msg = "Unable to insert document: check acdc server doc id: %s" % document['_id']
                raise RuntimeError(msg)
        return document

    @connectToCouch
    def listFiles(self):
        """
        _listFiles_

        return an iterator over the files contained in this fileset
        """
        for filelist in self.filelistDocuments():
            try:
                doc = self.couchdb.document(filelist)
            except CMSCouch.CouchNotFoundError as ex:
                msg = "Unable to retrieve Couch Document for fileset"
                msg += str(msg)
                raise RuntimeError(msg)

            files = doc["files"]
            for d in viewvalues(files):
                yield d

    @connectToCouch
    def fileset(self):
        """
        _fileset_

        Make a WMCore.DataStruct.Fileset instance containing the files in this fileset

        """
        result = DataStructsFileset(self['name'])
        pipeline = filePipeline({'fileset': result, 'run': makeRun({})})
        for f in self.listFiles():
            pipeline.send(f)
        return result

    @connectToCouch
    @requireFilesetName
    def populate(self):
        """
        _populate_

        Load all files out of couch.
        """
        params = {"startkey": [self.collectionName, self["name"]],
                  "endkey": [self.collectionName, self["name"]],
                  "include_docs": True, "reduce": False}
        result = self.couchdb.loadView("ACDC", "coll_fileset_docs",
                                       params)
        self.files = {}
        for row in result["rows"]:
            self.files.update(row["doc"]["files"])
            self["files"] = self.files
        return

    def fileCount(self):
        """
        _fileCount_

        Determine how many files are in the fileset.
        """
        params = {"startkey": [self.collectionName, self["name"]],
                  "endkey": [self.collectionName, self["name"]],
                  "reduce": True, "group_level": 2}
        result = self.couchdb.loadView("ACDC", "coll_fileset_count",
                                       params)
        return result["rows"][0]["value"]



#!/usr/bin/env python

"""
Defines default config values for JobAccountant specific
parameters.
"""
__all__ = []



from WMCore.Agent.Configuration import Configuration

config = Configuration()
config.component_("WorkQueueManager")

# set type of work queue
config.WorkQueueManager.level = "GlobalQueue"
#config.WorkQueueManager.level = "LocalQueue"

# In general can be left alone
config.WorkQueueManager.componentDir = config.General.WorkDir + "/WorkQueueManager"
config.WorkQueueManager.namespace = "WMComponent.WorkQueueManager.WorkQueueManager"
#config.WorkQueueManager.logLevel = 'INFO'
#config.WorkQueueManager.pollInterval = 600


# RequestManager config
config.WorkQueueManager.reqMgrConfig = {}
# uncomment to override default reqMgr url
#config.WorkQueueManager.reqMgrConfig['endpoint'] = 'http://cmssrv49.fnal.gov:8585'
config.WorkQueueManager.reqMgrConfig['teamName'] = 'Dodgers'


# add parameters for global or local queue if default param is not what you want
config.WorkQueueManager.queueParams = {'LocationRefreshInterval': 10}
# uncomment to change CacheDir from default
#config.WorkQueueManager.queueParams['CacheDir'] = os.path.join(config.WorkQueueManager.componentDir, 'wf')

# Fill for local queue
if config.WorkQueueManager.level != "GlobalQueue":
    # url for the global queue
    config.WorkQueueManager.queueParams['ParentQueue'] = 'http://example.com:8080/wq'
# used to identify & contact this queue.
# if not provided will attempt to get from REST configuration
#config.WorkQueueManager.queueParams['QueueURL'] = 'http://%s' % os.uname()[1]



#!/usr/bin/env python

"""
_Report_

Framework job report object.
"""
from __future__ import print_function
from builtins import str as newstr, bytes, range, object
from future.utils import viewitems, listitems

import logging
import math
import re
import sys
import time
import traceback

from Utils.PythonVersion import PY3
from Utils.Utilities import decodeBytesToUnicode, encodeUnicodeToBytes
from WMCore.Configuration import ConfigSection
from WMCore.DataStructs.File import File
from WMCore.DataStructs.Run import Run
from WMCore.FwkJobReport.FileInfo import FileInfo
from WMCore.WMException import WMException
from WMCore.WMExceptions import WM_JOB_ERROR_CODES

import pickle


class FwkJobReportException(WMException):
    """
    _FwkJobReportException_

    FwkJobReport error class
    """
    pass


def addBranchNamesToFile(fileSection, branchNames):
    """
    _addBranchNamesToFile_

    Given a list of branch names add them to the file section in the FWJR.
    """
    setattr(fileSection, "branches", branchNames)
    return


def addInputToFile(fileSection, inputLFN, inputPFN):
    """
    _addInputToFile_

    Given the LFN of an input file add it to the input section of the output
    file.
    """
    if not hasattr(fileSection, "input"):
        setattr(fileSection, "input", [])
    if not hasattr(fileSection, "inputpfns"):
        setattr(fileSection, "inputpfns", [])

    fileSection.input.append(inputLFN)
    fileSection.inputpfns.append(inputPFN)
    return


def addRunInfoToFile(fileSection, runInfo):
    """
    _addRunInfoToFile_

    Given a ConfigSection object corresponding to an input/output file in the
    FWJR and a WMCore.DataStructs.Run object add the run and lumi information
    to the ConfigSection.  It will have the following format:
      fileSection.runs.RUNNUMBER = [LUMI1, LUMI2, LUMI3...]

    Note that the run number will have to be cast to a string.
    """
    if not isinstance(runInfo, Run):
        for singleRun in runInfo:
            setattr(fileSection.runs, str(singleRun.run), singleRun.eventsPerLumi)
    else:
        setattr(fileSection.runs, str(runInfo.run), runInfo.eventsPerLumi)
    return


def addAttributesToFile(fileSection, **attributes):
    """
    _addAttributesToFiles_

    Add attributes to a file in the FWJR.
    """
    for attName, attValue in viewitems(attributes):
        setattr(fileSection, attName, attValue)
    return


class Report(object):
    """
    The base class for the new jobReport

    """

    def __init__(self, reportname=None):
        self.data = ConfigSection("FrameworkJobReport")
        self.data.steps = []
        self.data.workload = "Unknown"
        self.report = None
        self.reportname = ""

        if reportname:
            self.addStep(reportname=reportname)

        return

    def __str__(self):
        return str(self.data)

    def listSteps(self):
        """
        _listSteps_

        List the names of all the steps in the report.
        """
        return self.data.steps

    def setStepStatus(self, stepName, status):
        """
        _setStepStatus_

        Set the status for a step.
        """
        reportStep = self.retrieveStep(stepName)
        reportStep.status = status
        return

    def parse(self, xmlfile, stepName="cmsRun1"):
        """
        _parse_

        Read in the FrameworkJobReport XML file produced
        by cmsRun and pull the information from it into this object
        """
        from WMCore.FwkJobReport.XMLParser import xmlToJobReport
        try:
            xmlToJobReport(self, xmlfile)
        except Exception as ex:
            msg = "Error reading XML job report file, possibly corrupt XML File:\n"
            msg += "Details: %s" % str(ex)

            crashMessage = "\nStacktrace:\n"

            stackTrace = traceback.format_tb(sys.exc_info()[2], None)
            for stackFrame in stackTrace:
                crashMessage += stackFrame

            logging.debug(crashMessage)
            raise FwkJobReportException(msg)

    @staticmethod
    def jsonizeFiles(reportModule):
        """
        _jsonizeFiles_

        Put individual files in JSON format.
        """
        jsonFiles = []
        files = getattr(reportModule, "files", None)
        if not files:
            return jsonFiles

        fileCount = getattr(reportModule.files, "fileCount", 0)

        for i in range(fileCount):
            reportFile = getattr(reportModule.files, "file%s" % i)
            jsonFile = reportFile.dictionary_()

            if jsonFile.get('runs', None):
                cfgSectionRuns = jsonFile["runs"]
                jsonFile["runs"] = {}
                for runNumber in cfgSectionRuns.listSections_():
                    jsonFile["runs"][str(runNumber)] = getattr(cfgSectionRuns,
                                                               runNumber)
            jsonFiles.append(jsonFile)

        return jsonFiles

    @staticmethod
    def jsonizePerformance(perfSection):
        """
        _jsonizePerformance_

        Convert the performance section of the FWJR into JSON.
        """
        jsonPerformance = {}
        for reportSection in ["storage", "memory", "cpu", "multicore"]:
            jsonPerformance[reportSection] = {}
            if not hasattr(perfSection, reportSection):
                continue

            jsonPerformance[reportSection] = getattr(perfSection, reportSection).dictionary_()
            for key in jsonPerformance[reportSection]:
                val = jsonPerformance[reportSection][key]
                if isinstance(val, float):
                    if math.isinf(val) or math.isnan(val):
                        jsonPerformance[reportSection][key] = None

        return jsonPerformance

    def __to_json__(self, thunker):
        """
        __to_json__

        Create a JSON version of the Report.
        """
        jsonReport = {}
        jsonReport["WorkerNodeInfo"] = self.getWorkerNodeInfo()
        jsonReport["task"] = self.getTaskName()
        jsonReport["steps"] = {}
        jsonReport["skippedFiles"] = self.getAllSkippedFiles()
        jsonReport["fallbackFiles"] = self.getAllFallbackFiles()
        jsonReport["Campaign"] = self.getCampaign()
        jsonReport["PrepID"] = self.getPrepID()
        jsonReport["EOSLogURL"] = self.getLogURL()

        for stepName in self.listSteps():
            reportStep = self.retrieveStep(stepName)
            jsonStep = {}
            jsonStep["status"] = reportStep.status

            stepTimes = self.getTimes(stepName)

            if stepTimes["startTime"] is not None:
                stepTimes["startTime"] = int(stepTimes["startTime"])
            if stepTimes["stopTime"] is not None:
                stepTimes["stopTime"] = int(stepTimes["stopTime"])

            jsonStep["start"] = stepTimes["startTime"]
            jsonStep["stop"] = stepTimes["stopTime"]

            jsonStep["performance"] = self.jsonizePerformance(reportStep.performance)

            jsonStep["output"] = {}
            for outputModule in reportStep.outputModules:
                reportOutputModule = getattr(reportStep.output, outputModule)
                jsonStep["output"][outputModule] = self.jsonizeFiles(reportOutputModule)

            analysisSection = getattr(reportStep, 'analysis', None)
            if analysisSection:
                jsonStep["output"]['analysis'] = self.jsonizeFiles(analysisSection)

            jsonStep["input"] = {}
            for inputSource in reportStep.input.listSections_():
                reportInputSource = getattr(reportStep.input, inputSource)
                jsonStep["input"][inputSource] = self.jsonizeFiles(reportInputSource)

            jsonStep["errors"] = []
            errorCount = getattr(reportStep.errors, "errorCount", 0)
            for i in range(errorCount):
                reportError = getattr(reportStep.errors, "error%i" % i)
                jsonStep["errors"].append({"type": reportError.type,
                                           "details": reportError.details,
                                           "exitCode": reportError.exitCode})

            jsonStep["cleanup"] = {}
            jsonStep["parameters"] = {}
            jsonStep["site"] = self.getSiteName()
            jsonStep["analysis"] = {}
            jsonStep["logs"] = {}
            jsonReport["steps"][stepName] = jsonStep

        return jsonReport

    def getSiteName(self):
        """
        _getSiteName_

        Returns the site name attribute (no step specific)
        """
        return getattr(self.data, 'siteName', {})


    def _setSiteName(self, site):
        """
        _setSiteName_

        Set the site name attribute (no step specific)
        """
        setattr(self.data, 'siteName', site)

        return


    def getExitCodes(self):
        """
        _getExitCodes_

        Return a list of all non-zero exit codes in the report
        """
        returnCodes = set()
        for stepName in self.listSteps():
            returnCodes.update(self.getStepExitCodes(stepName=stepName))
        return returnCodes

    def getStepExitCodes(self, stepName):
        """
        _getStepExitCodes_

        Returns a list of all non-zero exit codes in the step
        """
        returnCodes = set()
        reportStep = self.retrieveStep(stepName)
        errorCount = getattr(reportStep.errors, "errorCount", 0)
        for i in range(errorCount):
            reportError = getattr(reportStep.errors, "error%i" % i)
            if getattr(reportError, 'exitCode', None):
                returnCodes.add(int(reportError.exitCode))
            else:
                # exitCode is likely set to None(?!?)
                returnCodes.add(99999)

        return returnCodes

    def getExitCode(self):
        """
        _getExitCode_

        Return the first exit code you find.
        """
        returnCode = 0
        for stepName in self.listSteps():
            errorCode = self.getStepExitCode(stepName=stepName)
            if errorCode == 99999:
                # Then we don't know what this error was
                # Mark it for return only if we don't fine an
                # actual error code in the job.
                returnCode = errorCode
            elif errorCode != 0:
                return errorCode

        return returnCode

    def getStepExitCode(self, stepName):
        """
        _getStepExitCode_

        Get the exit code for a particular step
        Return 0 if none
        """
        returnCode, _ = self.getStepExitCodeAndMessage(stepName)
        return returnCode

    def getStepExitCodeAndMessage(self, stepName):
        """
        _getStepExitCodeAndMessage_

        Get the exit code and message for a particular step
        Return (0, None)  if there were no errors.
        """
        returnCode = 0
        returnMessage = None
        reportStep = self.retrieveStep(stepName)
        errorCount = getattr(reportStep.errors, "errorCount", 0)
        for i in range(errorCount):
            reportError = getattr(reportStep.errors, "error%i" % i)
            if not getattr(reportError, 'exitCode', None):
                returnCode = 99999
                returnMessage = 'Unknown'
            else:
                return int(reportError.exitCode), reportError.details

        return returnCode, returnMessage

    def persist(self, filename):
        """
        _persist_

        Pickle this object and save it to disk.
        """
        if PY3:
            with open(filename, 'wb') as handle:
                pickle.dump(encodeUnicodeToBytes(self.data), handle)
        else:
            with open(filename, 'w') as handle:
                pickle.dump(self.data, handle)
        return

    def unpersist(self, filename, reportname=None):
        """
        _unpersist_

        Load a pickled FWJR from disk.
        """
        if PY3:
            with open(filename, 'rb') as handle:
                self.data = decodeBytesToUnicode(pickle.load(handle))
        else:
            with open(filename, 'r') as handle:
                self.data = pickle.load(handle)

        # old self.report (if it existed) became unattached
        if reportname:
            self.report = getattr(self.data, reportname)

        return

    def addOutputModule(self, moduleName):
        """
        _addOutputModule_

        Add an entry for an output module.
        """
        self.report.outputModules.append(moduleName)
        self.report.output.section_(moduleName)

        outMod = getattr(self.report.output, moduleName)
        outMod.section_("files")
        outMod.section_("dataset")
        outMod.files.fileCount = 0

        return outMod

    def killOutput(self):
        """
        _killOutput_

        Remove all the output from the report.  This is useful for chained
        processing where we don't want to keep the output from a particular
        step in a job.
        """
        for outputModuleName in self.report.outputModules:
            delattr(self.report.output, outputModuleName)

        self.report.outputModules = []
        return

    def addOutputFile(self, outputModule, aFile=None):
        """
        _addFile_

        Add an output file to the outputModule provided.
        """
        logging.info("addOutputFile method called with outputModule: %s, aFile: %s", outputModule, aFile)
        aFile = aFile or {}

        # Now load the output module and create the file object
        outMod = getattr(self.report.output, outputModule, None)
        if outMod is None:
            outMod = self.addOutputModule(outputModule)
        count = outMod.files.fileCount
        fileSection = "file%s" % count
        outMod.files.section_(fileSection)
        fileRef = getattr(outMod.files, fileSection)
        logging.info("addOutputFile method fileRef: %s, whole tree: %s", fileRef, fileRef.dictionary_whole_tree_())
        outMod.files.fileCount += 1

        # Now we need to eliminate the optional and non-primitives:
        # runs, parents, branches, locations and datasets
        keyList = list(aFile)

        fileRef.section_("runs")
        if "runs" in aFile:
            for run in aFile["runs"]:
                addRunInfoToFile(fileRef, run)
            keyList.remove('runs')

        if "parents" in aFile:
            setattr(fileRef, 'parents', list(aFile['parents']))
            keyList.remove('parents')

        if "locations" in aFile:
            fileRef.location = list(aFile["locations"])
            keyList.remove('locations')
        elif "PNN" in aFile:
            fileRef.location = [aFile["PNN"]]

        if "LFN" in aFile:
            fileRef.lfn = aFile["LFN"]
            keyList.remove("LFN")
        if "PFN" in aFile:
            fileRef.lfn = aFile["PFN"]
            keyList.remove("PFN")

        # All right, the rest should be JSONalizable python primitives
        for entry in keyList:
            setattr(fileRef, entry, aFile[entry])

        # And we're done
        return fileRef

    def addInputSource(self, sourceName):
        """
        _addInputSource_

        Add an input source to the report doing nothing if the input source
        already exists.
        """
        if hasattr(self.report.input, sourceName):
            return getattr(self.report.input, sourceName)

        self.report.input.section_(sourceName)
        srcMod = getattr(self.report.input, sourceName)
        srcMod.section_("files")
        srcMod.files.fileCount = 0

        return srcMod

    def addInputFile(self, sourceName, **attrs):
        """
        _addInputFile_

        Add an input file to the given source.
        """
        srcMod = getattr(self.report.input, sourceName, None)
        if srcMod is None:
            srcMod = self.addInputSource(sourceName)
        count = srcMod.files.fileCount
        fileSection = "file%s" % count
        srcMod.files.section_(fileSection)
        fileRef = getattr(srcMod.files, fileSection)
        srcMod.files.fileCount += 1

        keyList = list(attrs)

        fileRef.section_("runs")
        if "runs" in attrs:
            for run in attrs["runs"]:
                addRunInfoToFile(fileRef, run)
            keyList.remove('runs')

        if "parents" in attrs:
            keyList.remove('parents')
        if "locations" in attrs:
            keyList.remove('locations')

        # All right, the rest should be JSONalizable python primitives
        for entry in keyList:
            setattr(fileRef, entry, attrs[entry])

        return fileRef

    def addAnalysisFile(self, filename, **attrs):
        """
        _addAnalysisFile_

        Add an Analysis File.
        """
        analysisFiles = self.report.analysis.files
        count = analysisFiles.fileCount
        label = "file%s" % count

        analysisFiles.section_(label)
        newFile = getattr(analysisFiles, label)
        newFile.fileName = filename

        for x, y in viewitems(attrs):
            setattr(newFile, x, y)

        analysisFiles.fileCount += 1
        return

    def addRemovedCleanupFile(self, **attrs):
        """
        _addRemovedCleanupFile_

        Add a file to the cleanup.removed file
        """
        removedFiles = self.report.cleanup.removed
        count = self.report.cleanup.removed.fileCount
        label = 'file%s' % count

        removedFiles.section_(label)
        newFile = getattr(removedFiles, label)

        for x, y in viewitems(attrs):
            setattr(newFile, x, y)

        self.report.cleanup.removed.fileCount += 1
        return

    def addError(self, stepName, exitCode, errorType, errorDetails, siteName=None):
        """
        _addError_

        Add an error report with an exitCode, type/class of error and
        details of the error as a string.
        Also, report attempted site if error happened before landing on it.
        """
        if self.retrieveStep(stepName) is None:
            # Create a step and set it to failed
            # Assumption: Adding an error fails a step
            self.addStep(stepName, status=1)

        if exitCode is not None:
            exitCode = int(exitCode)

        setExitCodes = self.getStepExitCodes(stepName)
        if exitCode in setExitCodes:
            logging.warning("Exit code: %s has been already added to the job report", exitCode)
            return

        stepSection = self.retrieveStep(stepName)
        errorCount = getattr(stepSection.errors, "errorCount", 0)
        errEntry = "error%s" % errorCount
        stepSection.errors.section_(errEntry)
        errDetails = getattr(stepSection.errors, errEntry)
        errDetails.exitCode = exitCode
        errDetails.type = str(errorType)

        try:
            if isinstance(errorDetails, newstr):
                errDetails.details = errorDetails
            elif isinstance(errorDetails, bytes):
                errDetails.details = decodeBytesToUnicode(errorDetails, 'ignore')
            else:
                errDetails.details = newstr(errorDetails)
        except UnicodeEncodeError as ex:
            msg = "Failed to encode the job error details for job ID: %s." % self.getJobID()
            msg += "\nException message: %s\nOriginal error details: %s" % (str(ex), errorDetails)
            logging.error(msg)
            msg = "DEFAULT ERROR MESSAGE, because it failed to UTF-8 encode the original message."
            errDetails.details = msg
        except UnicodeDecodeError as ex:
            msg = "Failed to decode the job error details for job ID: %s." % self.getJobID()
            msg += "\nException message: %s\nOriginal error details: %s" % (str(ex), errorDetails)
            logging.error(msg)
            msg = "DEFAULT ERROR MESSAGE, because it failed to UTF-8 decode the original message."
            errDetails.details = msg

        setattr(stepSection.errors, "errorCount", errorCount + 1)
        self.setStepStatus(stepName=stepName, status=exitCode)

        if siteName:
            self._setSiteName(site=siteName)

        return

    def addSkippedFile(self, lfn, pfn):
        """
        _addSkippedFile_

        Report a skipped input file
        """
        count = self.report.skipped.files.fileCount
        entry = "file%s" % count
        self.report.skipped.files.section_(entry)
        skipSect = getattr(self.report.skipped.files, entry)
        skipSect.PhysicalFileName = pfn
        skipSect.LogicalFileName = lfn
        self.report.skipped.files.fileCount += 1
        return

    def addFallbackFile(self, lfn, pfn):
        """
        _addFallbackFile_

        Report a fallback attempt for input file
        """
        count = self.report.fallback.files.fileCount
        entry = "file%s" % count
        self.report.fallback.files.section_(entry)
        fallbackSect = getattr(self.report.fallback.files, entry)
        fallbackSect.PhysicalFileName = pfn
        fallbackSect.LogicalFileName = lfn
        self.report.fallback.files.fileCount += 1
        return

    def addSkippedEvent(self, run, event):
        """
        _addSkippedEvent_

        Add a skipped event.
        """
        self.report.skipped.events.section_(str(run))
        runsect = getattr(self.report.skipped.events, str(run))
        if not hasattr(runsect, "eventList"):
            runsect.eventList = []
        runsect.eventList.append(event)
        return

    def addStep(self, reportname, status=1):
        """
        _addStep_

        This creates a report section into self.report
        """
        if hasattr(self.data, reportname):
            msg = "Attempted to create pre-existing report section %s" % reportname
            logging.error(msg)
            return

        self.data.steps.append(reportname)

        self.reportname = reportname
        self.data.section_(reportname)
        self.report = getattr(self.data, reportname)
        self.report.id = None
        self.report.status = status
        self.report.outputModules = []

        # structure
        self.report.section_("site")
        self.report.section_("output")
        self.report.section_("input")
        self.report.section_("performance")
        self.report.section_("analysis")
        self.report.section_("errors")
        self.report.section_("skipped")
        self.report.section_("fallback")
        self.report.section_("parameters")
        self.report.section_("logs")
        self.report.section_("cleanup")
        self.report.analysis.section_("files")
        self.report.cleanup.section_("removed")
        self.report.cleanup.section_("unremoved")
        self.report.skipped.section_("events")
        self.report.skipped.section_("files")
        self.report.fallback.section_("files")
        self.report.skipped.files.fileCount = 0
        self.report.fallback.files.fileCount = 0
        self.report.analysis.files.fileCount = 0
        self.report.cleanup.removed.fileCount = 0

        return

    def setStep(self, stepName, stepSection):
        """
        _setStep_

        """
        if stepName not in self.listSteps():
            self.data.steps.append(stepName)
        else:
            logging.info("Step %s is now being overridden by a new step report", stepName)
        self.data.section_(stepName)
        setattr(self.data, stepName, stepSection)
        return

    def retrieveStep(self, step):
        """
        _retrieveStep_

        Grabs a report in the raw and returns it.
        """
        reportSection = getattr(self.data, step, None)
        return reportSection

    def load(self, filename):
        """
        _load_

        This just maps to unpersist
        """
        self.unpersist(filename)
        return

    def save(self, filename):
        """
        _save_

        This just maps to persist
        """
        self.persist(filename)
        return

    def getOutputModule(self, step, outputModule):
        """
        _getOutputModule_

        Get the output module from a particular step
        """
        stepReport = self.retrieveStep(step=step)

        if not stepReport:
            return None

        return getattr(stepReport.output, outputModule, None)

    def getOutputFile(self, fileName, outputModule, step):
        """
        _getOutputFile_

        Takes a fileRef object and returns a DataStructs/File object as output
        """

        outputMod = self.getOutputModule(step=step, outputModule=outputModule)

        if not outputMod:
            return None

        fileRef = getattr(outputMod.files, fileName, None)
        newFile = File(locations=set())

        # Locations
        newFile.setLocation(getattr(fileRef, "location", None))

        # Runs
        runList = fileRef.runs.listSections_()
        for run in runList:
            lumis = getattr(fileRef.runs, run)
            if isinstance(lumis, dict):
                newRun = Run(int(run), *listitems(lumis))
            else:
                newRun = Run(int(run), *lumis)
            newFile.addRun(newRun)

        newFile["lfn"] = getattr(fileRef, "lfn", None)
        newFile["pfn"] = getattr(fileRef, "pfn", None)
        newFile["events"] = int(getattr(fileRef, "events", 0))
        newFile["size"] = int(getattr(fileRef, "size", 0))
        newFile["branches"] = getattr(fileRef, "branches", [])
        newFile["input"] = getattr(fileRef, "input", [])
        newFile["inputpfns"] = getattr(fileRef, "inputpfns", [])
        newFile["branch_hash"] = getattr(fileRef, "branch_hash", None)
        newFile["catalog"] = getattr(fileRef, "catalog", "")
        newFile["guid"] = getattr(fileRef, "guid", "")
        newFile["module_label"] = getattr(fileRef, "module_label", "")
        newFile["checksums"] = getattr(fileRef, "checksums", {})
        newFile["merged"] = bool(getattr(fileRef, "merged", False))
        newFile["dataset"] = getattr(fileRef, "dataset", {})
        newFile["acquisitionEra"] = getattr(fileRef, 'acquisitionEra', None)
        newFile["processingVer"] = getattr(fileRef, 'processingVer', 0)
        newFile["validStatus"] = getattr(fileRef, 'validStatus', None)
        newFile["globalTag"] = getattr(fileRef, 'globalTag', None)
        newFile["prep_id"] = getattr(fileRef, 'prep_id', None)
        newFile['configURL'] = getattr(fileRef, 'configURL', None)
        newFile['inputPath'] = getattr(fileRef, 'inputPath', None)
        newFile["outputModule"] = outputModule
        newFile["fileRef"] = fileRef

        return newFile

    def getAllFilesFromStep(self, step):
        """
        _getAllFilesFromStep_

        For a given step, retrieve all the associated files
        """

        stepReport = self.retrieveStep(step=step)
        if not stepReport:
            logging.debug("Asked to retrieve files from non-existant step %s", step)
            return []

        # steps with no outputModules can be ok (even for CMSSW steps)
        listOfModules = getattr(stepReport, 'outputModules', None)
        if not listOfModules:
            return []

        listOfFiles = []
        for module in listOfModules:
            listOfFiles.extend(self.getFilesFromOutputModule(step=step, outputModule=module))

        return listOfFiles

    def getAllFiles(self):
        """
        _getAllFiles_

        Grabs all files in all output modules in all steps
        """
        listOfFiles = []

        for step in self.listSteps():
            listOfFiles.extend(self.getAllFilesFromStep(step=step))

        return listOfFiles

    def getAllInputFiles(self):
        """
        _getAllInputFiles_

        Gets all the input files
        """

        listOfFiles = []
        for step in self.listSteps():
            tmp = self.getInputFilesFromStep(stepName=step)
            if tmp:
                listOfFiles.extend(tmp)

        return listOfFiles

    def getInputFilesFromStep(self, stepName, inputSource=None):
        """
        _getInputFilesFromStep_

        Retrieve a list of input files from the given step.
        """
        step = self.retrieveStep(stepName)

        if inputSource is None:
            inputSources = step.input.listSections_()
        else:
            inputSources = [inputSource]

        inputFiles = []
        for inputSource in inputSources:
            source = getattr(step.input, inputSource)
            for fileNum in range(source.files.fileCount):
                fwjrFile = getattr(source.files, "file%d" % fileNum)

                lfn = getattr(fwjrFile, "lfn", None)
                pfn = getattr(fwjrFile, "pfn", None)
                size = getattr(fwjrFile, "size", 0)
                events = getattr(fwjrFile, "events", 0)
                branches = getattr(fwjrFile, "branches", [])
                catalog = getattr(fwjrFile, "catalog", None)
                guid = getattr(fwjrFile, "guid", None)
                inputSourceClass = getattr(fwjrFile, "input_source_class", None)
                moduleLabel = getattr(fwjrFile, "module_label", None)
                inputType = getattr(fwjrFile, "input_type", None)

                inputFile = File(lfn=lfn, size=size, events=events)
                inputFile["pfn"] = pfn
                inputFile["branches"] = branches
                inputFile["catalog"] = catalog
                inputFile["guid"] = guid
                inputFile["input_source_class"] = inputSourceClass
                inputFile["module_label"] = moduleLabel
                inputFile["input_type"] = inputType

                runSection = getattr(fwjrFile, "runs")
                runNumbers = runSection.listSections_()

                for runNumber in runNumbers:
                    lumiTuple = getattr(runSection, str(runNumber))
                    inputFile.addRun(Run(int(runNumber), *lumiTuple))

                inputFiles.append(inputFile)

        return inputFiles

    def getFilesFromOutputModule(self, step, outputModule):
        """
        _getFilesFromOutputModule_

        Grab all the files in a particular output module
        """

        outputMod = self.getOutputModule(step=step, outputModule=outputModule)

        if not outputMod:
            return []

        listOfFiles = []
        for n in range(outputMod.files.fileCount):
            aFile = self.getOutputFile(fileName='file%i' % (n), outputModule=outputModule, step=step)
            if aFile:
                listOfFiles.append(aFile)
            else:
                msg = "Could not find file%i in module" % (n)
                logging.error(msg)
                return []

        return listOfFiles

    def getAllSkippedFiles(self):
        """
        _getAllSkippedFiles_

        Get a list of LFNs for all the input files
        listed as skipped on the report.
        """
        listOfFiles = []
        for step in self.listSteps():
            tmp = self.getSkippedFilesFromStep(stepName=step)
            if tmp:
                listOfFiles.extend(tmp)

        return listOfFiles

    def getAllFallbackFiles(self):
        """
        _getAllFallbackFiles_

        Get a list of LFNs for all the input files
        listed as fallback attempt on the report
        """
        listOfFiles = []
        for step in self.listSteps():
            tmp = self.getFallbackFilesFromStep(stepName=step)
            if tmp:
                listOfFiles.extend(tmp)

        return listOfFiles

    def getSkippedFilesFromStep(self, stepName):
        """
        _getSkippedFilesFromStep_

        Get a list of LFNs skipped in the given step
        """
        skippedFiles = []

        step = self.retrieveStep(stepName)

        filesSection = step.skipped.files
        fileCount = getattr(filesSection, "fileCount", 0)

        for fileNum in range(fileCount):
            fileSection = getattr(filesSection, "file%d" % fileNum)
            lfn = getattr(fileSection, "LogicalFileName", None)
            if lfn is not None:
                skippedFiles.append(lfn)
            else:
                logging.error("Found no LFN in file %s", str(fileSection))

        return skippedFiles

    def getFallbackFilesFromStep(self, stepName):
        """
        _getFallbackFilesFromStep_

        Get a list of LFNs which triggered a fallback in the given step
        """
        fallbackFiles = []

        step = self.retrieveStep(stepName)
        try:
            filesSection = step.fallback.files
        except AttributeError:
            return fallbackFiles
        fileCount = getattr(filesSection, "fileCount", 0)

        for fileNum in range(fileCount):
            fileSection = getattr(filesSection, "file%d" % fileNum)
            lfn = getattr(fileSection, "LogicalFileName", None)
            if lfn is not None:
                fallbackFiles.append(lfn)
            else:
                logging.error("Found no LFN in file %s", str(fileSection))

        return fallbackFiles

    def getStepErrors(self, stepName):
        """
        _getStepErrors_

        Get all errors for a given step
        """
        if self.retrieveStep(stepName) is None:
            # Create a step and set it to failed
            # Assumption: Adding an error fails a step
            self.addStep(stepName, status=1)

        stepSection = self.retrieveStep(stepName)

        errorCount = getattr(stepSection.errors, "errorCount", 0)
        if errorCount == 0:
            return {}
        else:
            return stepSection.errors.dictionary_()

    def stepSuccessful(self, stepName):
        """
        _stepSuccessful_

        Determine wether or not a step was successful.
        """
        stepReport = self.retrieveStep(step=stepName)
        status = getattr(stepReport, 'status', 1)
        # We have too many possibilities
        if status not in [0, '0', 'success', 'Success']:
            return False

        return True

    def taskSuccessful(self, ignoreString='logArch'):
        """
        _taskSuccessful_

        Return True if all steps successful, False otherwise
        """
        value = True

        if len(self.listSteps()) == 0:
            # Mark jobs as failed if they have no steps
            msg = "Could not find any steps"
            logging.error(msg)
            return False

        for stepName in self.listSteps():
            # Ignore specified steps
            # i.e., logArch steps can fail without causing
            # the task to fail
            if ignoreString and re.search(ignoreString, stepName):
                continue
            if not self.stepSuccessful(stepName=stepName):
                value = False

        return value

    def getAnalysisFilesFromStep(self, step):
        """
        _getAnalysisFilesFromStep_

        Retrieve a list of all the analysis files produced in a step.
        """
        stepReport = self.retrieveStep(step=step)

        if not stepReport or not hasattr(stepReport.analysis, 'files'):
            return []

        analysisFiles = stepReport.analysis.files

        results = []
        for fileNum in range(analysisFiles.fileCount):
            results.append(getattr(analysisFiles, "file%s" % fileNum))

        # filter out duplicates
        duplicateCheck = []
        filteredResults = []
        for result in results:
            inputtag = getattr(result, 'inputtag', None)
            if (result.fileName, inputtag) not in duplicateCheck:
                duplicateCheck.append((result.fileName, inputtag))
                filteredResults.append(result)

        return filteredResults

    def getAllFileRefsFromStep(self, step):
        """
        _getAllFileRefsFromStep_

        Retrieve a list of all files produced in a step.  The files will be in
        the form of references to the ConfigSection objects in the acutal
        report.
        """
        stepReport = self.retrieveStep(step=step)
        if not stepReport:
            return []

        outputModules = getattr(stepReport, "outputModules", [])
        fileRefs = []
        for outputModule in outputModules:
            outputModuleRef = self.getOutputModule(step=step, outputModule=outputModule)

            for i in range(outputModuleRef.files.fileCount):
                fileRefs.append(getattr(outputModuleRef.files, "file%i" % i))

        analysisFiles = self.getAnalysisFilesFromStep(step)
        fileRefs.extend(analysisFiles)

        return fileRefs

    def addInfoToOutputFilesForStep(self, stepName, step):
        """
        _addInfoToOutputFilesForStep_

        Add the information missing from output files to the files
        This requires the WMStep to be passed in
        """

        stepReport = self.retrieveStep(step=stepName)
        fileInfo = FileInfo()

        if not stepReport:
            return

        listOfModules = getattr(stepReport, 'outputModules', None)

        for module in listOfModules:
            outputMod = getattr(stepReport.output, module, None)
            for n in range(outputMod.files.fileCount):
                aFile = getattr(outputMod.files, 'file%i' % n, None)
                if not aFile:
                    msg = "Could not find file%i in module" % n
                    logging.error(msg)
                    return
                fileInfo(fileReport=aFile, step=step, outputModule=module)

        return

    def deleteOutputModuleForStep(self, stepName, moduleName):
        """
        _deleteOutputModuleForStep_

        Delete any reference to the given output module in the step report
        that includes deleting any output file it produced
        """
        stepReport = self.retrieveStep(step=stepName)

        if not stepReport:
            return

        listOfModules = getattr(stepReport, 'outputModules', [])

        if moduleName not in listOfModules:
            return

        delattr(stepReport.output, moduleName)
        listOfModules.remove(moduleName)

        return

    def setStepStartTime(self, stepName):
        """
        _setStepStatus_

        Set the startTime for a step.
        """
        reportStep = self.retrieveStep(stepName)
        reportStep.startTime = time.time()
        return

    def setStepStopTime(self, stepName):
        """
        _setStepStatus_

        Set the stopTime for a step.
        """
        reportStep = self.retrieveStep(stepName)
        reportStep.stopTime = time.time()
        return

    def getTimes(self, stepName):
        """
        _getTimes_

        Return a dictionary with the start and stop times
        """
        reportStep = self.retrieveStep(stepName)

        startTime = getattr(reportStep, 'startTime', None)
        stopTime = getattr(reportStep, 'stopTime', None)

        return {'startTime': startTime, 'stopTime': stopTime}

    def getFirstStartLastStop(self):
        """
        _getFirstStartLastStop_

        Get the first startTime, last stopTime
        """

        steps = self.listSteps()

        if len(steps) < 1:
            return None

        firstStep = self.getTimes(stepName=steps[0])
        startTime = firstStep['startTime']
        stopTime = firstStep['stopTime']

        for stepName in steps:
            timeStamps = self.getTimes(stepName=stepName)
            if timeStamps['startTime'] is None or timeStamps['stopTime'] is None:
                # Unusable times
                continue
            if startTime is None or startTime > timeStamps['startTime']:
                startTime = timeStamps['startTime']
            if stopTime is None or stopTime < timeStamps['stopTime']:
                stopTime = timeStamps['stopTime']

        return {'startTime': startTime, 'stopTime': stopTime}

    def setTaskName(self, taskName):
        """
        _setTaskName_

        Set the task name for the report
        """
        self.data.task = taskName
        return

    def getTaskName(self):
        """
        _getTaskName_

        Return the task name
        """
        return getattr(self.data, 'task', None)

    def setJobID(self, jobID):
        """
        _setJobID_

        Set the WMBS jobID
        """

        self.data.jobID = jobID
        return

    def getJobID(self):
        """
        _getJobID_

        Get the WMBS job ID if attached
        """

        return getattr(self.data, 'jobID', None)

    def getAllFileRefs(self):
        """
        _getAllFileRefs_

        Get references for all file in the step
        """

        fileRefs = []
        for step in self.listSteps():
            tmpRefs = self.getAllFileRefsFromStep(step=step)
            if len(tmpRefs) > 0:
                fileRefs.extend(tmpRefs)

        return fileRefs

    def setAcquisitionProcessing(self, acquisitionEra, processingVer, processingStr=None):
        """
        _setAcquisitionProcessing_

        Set the acquisition and processing era for every output file
        ONLY run this after all files have been accumulated; it doesn't
        set things for future files.
        """
        fileRefs = self.getAllFileRefs()

        # Should now have all the fileRefs
        for f in fileRefs:
            f.acquisitionEra = acquisitionEra
            f.processingVer = processingVer
            f.processingStr = processingStr

        return

    def setValidStatus(self, validStatus):
        """
        _setValidStatus_

        Set the validStatus for all steps and all files.
        ONLY run this after all files have been attached.
        """

        fileRefs = self.getAllFileRefs()

        # Should now have all the fileRefs
        for f in fileRefs:
            f.validStatus = validStatus

        return

    def setGlobalTag(self, globalTag):
        """
        _setGlobalTag_

        Set the global Tag from the spec on the WN
        ONLY run this after all the files have been attached
        """

        fileRefs = self.getAllFileRefs()

        # Should now have all the fileRefs
        for f in fileRefs:
            f.globalTag = globalTag

        return

    def setCampaign(self, campaign):
        """
        _setCampaign_
        Set the campaign for the report
        """
        self.data.campaign = campaign
        return

    def getCampaign(self):
        """
        _getCampaign_
        Return the campaign
        """
        return getattr(self.data, 'campaign', "")

    def setPrepID(self, prep_id):
        """
        _setGlobalTag_

        Set the global Tag from the spec on the WN
        ONLY run this after all the files have been attached
        """

        fileRefs = self.getAllFileRefs()

        # Should now have all the fileRefs
        for f in fileRefs:
            f.prep_id = prep_id

        self.data.prep_id = prep_id
        return

    def getPrepID(self):
        """
         _getPrepID_

         Return the PrepID
        """
        return getattr(self.data, 'prep_id', "")

    def setConfigURL(self, configURL):
        """
        _setConfigURL_

        Set the config URL in a portable storage form
        """

        fileRefs = self.getAllFileRefs()

        # Should now have all the fileRefs
        for f in fileRefs:
            f.configURL = configURL

        return

    def setInputDataset(self, inputPath):
        """
        _setInputDataset_

        Set the input dataset path for the task in each file
        """

        fileRefs = self.getAllFileRefs()

        # Should now have all the fileRefs
        for f in fileRefs:
            f.inputPath = inputPath
        return

    def setStepRSS(self, stepName, minimum, maximum, average):
        """
        _setStepRSS_

        Set the Performance RSS information
        """

        reportStep = self.retrieveStep(stepName)
        reportStep.performance.section_('RSSMemory')
        reportStep.performance.RSSMemory.min = minimum
        reportStep.performance.RSSMemory.max = maximum
        reportStep.performance.RSSMemory.average = average

        return

    def setStepPMEM(self, stepName, minimum, maximum, average):
        """
        _setStepPMEM_

        Set the Performance PMEM information
        """

        reportStep = self.retrieveStep(stepName)
        reportStep.performance.section_('PhysicalMemory')
        reportStep.performance.PhysicalMemory.min = minimum
        reportStep.performance.PhysicalMemory.max = maximum
        reportStep.performance.PhysicalMemory.average = average

        return

    def setStepPCPU(self, stepName, minimum, maximum, average):
        """
        _setStepPCPU_

        Set the Performance PCPU information
        """

        reportStep = self.retrieveStep(stepName)
        reportStep.performance.section_('PercentCPU')
        reportStep.performance.PercentCPU.min = minimum
        reportStep.performance.PercentCPU.max = maximum
        reportStep.performance.PercentCPU.average = average

        return

    def setStepVSize(self, stepName, minimum, maximum, average):
        """
        _setStepVSize_

        Set the Performance PCPU information
        """

        reportStep = self.retrieveStep(stepName)
        reportStep.performance.section_('VSizeMemory')
        reportStep.performance.VSizeMemory.min = minimum
        reportStep.performance.VSizeMemory.max = maximum
        reportStep.performance.VSizeMemory.average = average

        return

    def setStepCounter(self, stepName, counter):
        """
        _setStepCounter_

        Assign a number to the step
        """

        reportStep = self.retrieveStep(stepName)
        reportStep.counter = counter

        return

    def checkForAdlerChecksum(self, stepName):
        """
        _checkForAdlerChecksum_

        Some steps require that all output files have adler checksums
        This will go through all output files in a step and make sure they
          have an adler32 checksum.  If they don't it creates an error with
          code 60451 for the step, failing it.
        """
        error = None
        files = self.getAllFilesFromStep(step=stepName)
        for f in files:
            if 'adler32' not in f.get('checksums', {}):
                error = f.get('lfn', None)
            elif f['checksums']['adler32'] is None:
                error = f.get('lfn', None)

        if error:
            msg = '%s, file was %s' % (WM_JOB_ERROR_CODES[60451], error)
            self.addError(stepName, 60451, "NoAdler32Checksum", msg)
            self.setStepStatus(stepName=stepName, status=60451)

        return

    def checkForRunLumiInformation(self, stepName):
        """
        _checkForRunLumiInformation_

        Some steps require that all output files have run lumi information.
        This will go through all output files in a step and make sure
        they have run/lumi informaiton. If they don't it creates an error
        with code 70452 for the step, failing it.

        """
        error = None
        files = self.getAllFilesFromStep(step=stepName)
        for f in files:
            if not f.get('runs', None):
                error = f.get('lfn', None)
            else:
                for run in f['runs']:
                    lumis = run.lumis
                    if not lumis:
                        error = f.get('lfn', None)
                        break
        if error:
            msg = '%s, file was %s' % (WM_JOB_ERROR_CODES[70452], error)
            self.addError(stepName, 70452, "NoRunLumiInformation", msg)
            self.setStepStatus(stepName=stepName, status=70452)
        return

    def checkForOutputFiles(self, stepName):
        """
        _checkForOutputFiles_

        Verify that there is at least an output file, either from
        analysis or from an output module.
        """
        files = self.getAllFilesFromStep(step=stepName)
        analysisFiles = self.getAnalysisFilesFromStep(step=stepName)
        if len(files) == 0 and len(analysisFiles) == 0:
            msg = WM_JOB_ERROR_CODES[60450]
            msg += "\nList of skipped files is:\n"
            for skipF in self.getSkippedFilesFromStep(stepName=stepName):
                msg += "  %s\n" % skipF
            self.addError(stepName, 60450, "NoOutput", msg)
            self.setStepStatus(stepName=stepName, status=60450)
        return

    def stripInputFiles(self):
        """
        _stripInputFiles_

        If we need to compact the FWJR the easiest way is just to
        trim the number of input files.
        """

        for stepName in self.listSteps():
            step = self.retrieveStep(stepName)
            inputSources = step.input.listSections_()
            for inputSource in inputSources:
                source = getattr(step.input, inputSource)
                for fileNum in range(source.files.fileCount):
                    delattr(source.files, "file%d" % fileNum)
                source.files.fileCount = 0
        return

    def getWorkerNodeInfo(self):
        wnInfo = {"HostName": getattr(self.data, 'hostName', ''),
                  "MachineFeatures": getattr(self.data, 'machineFeatures', {}),
                  "JobFeatures": getattr(self.data, 'jobFeatures', {})}

        return wnInfo

    def setLogURL(self, url):
        """
        Set log url for the this job report.
        https://eoscmsweb.cern.ch/eos/cms/store/logs/prod/recent/
        """
        self.data.logURL = url

    def getLogURL(self):
        """
        _getLogURL_
        Return the log URL
        """
        return getattr(self.data, 'logURL', '')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Script to connect to a MongoDB instance based on a set of parameters given at runtime
Usage:
      source /data/srv/current/apps/reqmgr2ms/etc/profile.d/init.sh
      python3 -i -- mongoInit.py -c $WMCORE_SERVICE_CONFIG/reqmgr2ms-output/config-output.py
"""

import sys
import logging
import argparse
from pprint import pformat

from WMCore.Database.MongoDB import MongoDB

from WMCore.Configuration import loadConfigurationFile

if __name__ == '__main__':

    FORMAT = "%(asctime)s:%(levelname)s:%(module)s:%(funcName)s(): %(message)s"
    logging.basicConfig(stream=sys.stdout, format=FORMAT, level=logging.INFO)
    logger = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(
        prog='MongoInit',
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__)

    parser.add_argument('-c', '--config', required=True,
                        help="""\
                        The path to MSCONFIG to be used, e.g.
                        for production: /data/srv/current/config/reqmgr2ms/config-output.py
                        """)
    args = parser.parse_args()

    logger.info("Loading configFile: %s", args.config)
    config = loadConfigurationFile(args.config)

    msConfig = config.section_('views').section_('data').dictionary_()

    msConfig.setdefault("mongoDBUrl", 'mongodb://localhost')
    msConfig.setdefault("mongoDBPort", None)
    msConfig.setdefault("mongoDB", 'mongoDBNull')
    msConfig.setdefault("mongoDBRetryCount", 3)
    msConfig.setdefault("mongoDBReplicaSet", None)
    msConfig.setdefault("mockMongoDB", False)
    msConfig.setdefault("collection", None)

    # NOTE: A full set of valid database connection parameters can be found at:
    #       https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html
    mongoDBConfig = {
        'database': msConfig['mongoDB'],
        'server': msConfig['mongoDBServer'],
        'replicaSet': msConfig['mongoDBReplicaSet'],
        'port': msConfig['mongoDBPort'],
        'username': msConfig['mongoDBUser'],
        'password': msConfig['mongoDBPassword'],
        'connect': True,
        'directConnection': False,
        'logger': logger,
        'create': False,
        'mockMongoDB': msConfig['mockMongoDB']}

    # NOTE: We need to blur `username' and `password' keys before printing the configuration:
    msg = "Connecting to MongoDB using the following mongoDBConfig:\n%s"
    logger.info(msg, pformat({**mongoDBConfig, **{'username': '****', 'password': '****'}}))

    mongoDB = MongoDB(**mongoDBConfig)
    currDB = getattr(mongoDB, msConfig['mongoDB'])
    mongoClt = mongoDB.client
    mongoColl = currDB[msConfig['collection']] if msConfig['collection'] else None



#!/usr/bin/env python
"""
_ListSites_

Oracle implementation of Locations.ListSites
"""

__all__ = []



from WMCore.WMBS.MySQL.Locations.ListSites import ListSites as MySQLListSites

class ListSites(MySQLListSites):
    """
    Right now the same as the MySQL implementation

    """



"""
Patterns module provides set of CS patterns
"""
from __future__ import division, print_function


class Singleton(type):
    """Implementation of Singleton class"""
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = \
                    super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from __future__ import (division, print_function)
import time
from pprint import pformat
from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.Services.WMStats.WMStatsWriter import WMStatsWriter, convertToServiceCouchDoc

# CMSMonitoring modules
from CMSMonitoring.StompAMQ7 import StompAMQ7 as StompAMQ


class HeartbeatMonitorBase(CherryPyPeriodicTask):

    def __init__(self, rest, config):
        super(HeartbeatMonitorBase, self).__init__(config)
        self.centralWMStats = WMStatsWriter(config.wmstats_url)
        self.threadList = config.thread_list
        self.userAMQ = getattr(config, "user_amq", None)
        self.passAMQ = getattr(config, "pass_amq", None)
        self.postToAMQ = getattr(config, "post_to_amq", False)
        self.topicAMQ = getattr(config, "topic_amq", None)
        self.hostPortAMQ = getattr(config, "host_port_amq", None)

    def setConcurrentTasks(self, config):
        """
        sets the list of function reference for concurrent tasks
        """
        self.concurrentTasks = [{'func': self.reportToWMStats, 'duration': config.heartbeatCheckDuration}]

    def reportToWMStats(self, config):
        """
        report thread status and heartbeat.
        Also can report additional monitoring information by rewriting addAdditionalMonitorReport method
        """
        self.logger.info("Checking Thread status...")
        downThreadInfo = self.logDB.wmstats_down_components_report(self.threadList)
        monitorInfo = self.addAdditionalMonitorReport(config)
        downThreadInfo.update(monitorInfo)
        wqSummaryDoc = convertToServiceCouchDoc(downThreadInfo, config.log_reporter)
        self.centralWMStats.updateAgentInfo(wqSummaryDoc)

        self.logger.info("Uploaded to WMStats...")

        return

    def addAdditionalMonitorReport(self, config):
        """
        add Additonal report with heartbeat report
        overwite the method with each applications monitoring info. (Need to follow the format displayed in wmstats)
        """
        return {}

    def uploadToAMQ(self, docs, producer=None):
        """
        _uploadToAMQ_

        Sends data to AMQ, which ends up in elastic search.
        :param docs: list of documents/dicts to be posted
        :param producer: service name that's providing this info
        """
        if not docs:
            self.logger.info("There are no documents to send to AMQ")
            return

        producer = producer or self.producer
        ts = int(time.time())
        notifications = []

        self.logger.debug("Sending the following data to AMQ %s", pformat(docs))
        try:
            stompSvc = StompAMQ(username=self.userAMQ,
                                password=self.passAMQ,
                                producer=producer,
                                topic=self.topicAMQ,
                                validation_schema=None,
                                host_and_ports=self.hostPortAMQ,
                                logger=self.logger)

            for doc in docs:
                singleNotif, _, _ = stompSvc.make_notification(payload=doc, doc_type=self.docTypeAMQ,
                                                               ts=ts, data_subfield="payload")
                notifications.append(singleNotif)

            failures = stompSvc.send(notifications)
            msg = "%i out of %i documents successfully sent to AMQ" % (len(notifications) - len(failures),
                                                                       len(notifications))
            self.logger.info(msg)
        except Exception as ex:
            self.logger.exception("Failed to send data to StompAMQ. Error %s", str(ex))

        return



from __future__ import (print_function, division)
from builtins import object

class DataCache(object):
    _finishedWFs = {} # Global cache holding workflow lists

    @staticmethod
    def getFinishedWorkflows():
        return DataCache._finishedWFs

    @staticmethod
    def setFinishedWorkflows(workflowLists):
        DataCache._finishedWFs = workflowLists



#!/usr/bin/env python
# pylint: disable=C0321,C0103,W0622
# W0622: redefined-builtin
"""
_Configuration_

Module dealing with Configuration file in python format


"""

from builtins import object, int, str as newstr, bytes as newbytes

from future.utils import listvalues

import os
import traceback

from Utils.PythonVersion import PY3

import importlib

_SimpleTypes = [
    bool,
    float,
    # basestring,  # For py2/py3 compatibility, don't let futurize remove PY3 Remove when python 3 transition complete
    newbytes,
    newstr,
    type(None),
    int,
]

_ComplexTypes = [
    dict,
    list,
    tuple,
]

_SupportedTypes = []
_SupportedTypes.extend(_SimpleTypes)
_SupportedTypes.extend(_ComplexTypes)


def formatAsString(value):
    """
    _format_

    format a value as python
    keep parameters simple, trust python...
    """
    if isinstance(value, str):
        value = "\'%s\'" % value
    return str(value)


def formatNative(value):
    """
    _formatNative_

    Like the format function, but allowing passing of ints, floats, etc.
    """

    if isinstance(value, int):
        return value
    if isinstance(value, float):
        return value
    if isinstance(value, list):
        return value
    if isinstance(value, dict):
        return dict
    return formatAsString(value)


class ConfigSection(object):
    # pylint: disable=protected-access
    """
    _ConfigSection_

    Chunk of configuration information
    """

    def __init__(self, name=None):
        object.__init__(self)
        self._internal_documentation = ""
        self._internal_name = name
        self._internal_settings = set()
        self._internal_docstrings = {}
        self._internal_children = set()
        self._internal_parent_ref = None
        # The flag skipChecks controls weather each parameter added to the configuration
        # should be a primitive or complex type that can be "jsonized"
        # By default it is false, but ConfigurationEx instances set it to True
        self._internal_skipChecks = False

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                (self._internal_documentation == other._internal_documentation) and
                (self._internal_name == other._internal_name) and
                (self._internal_settings == other._internal_settings) and
                (self._internal_docstrings == other._internal_docstrings) and
                (self._internal_children == other._internal_children) and
                (self._internal_parent_ref == other._internal_parent_ref))
        return id(self) == id(other)

    def _complexTypeCheck(self, name, value):

        if isinstance(value, tuple(_SimpleTypes)):
            return
        elif isinstance(value, tuple(_ComplexTypes)):
            vallist = value
            if isinstance(value, dict):
                vallist = listvalues(value)
            for val in vallist:
                self._complexTypeCheck(name, val)
        else:
            msg = "Not supported type in sequence:"
            msg += "%s\n" % type(value)
            msg += "for name: %s and value: %s\n" % (name, value)
            msg += "Added to WMAgent Configuration."
            msg += "Use ConfigurationEx to skip checks on config params"
            raise RuntimeError(msg)

    def __setattr__(self, name, value):
        if name.startswith("_internal_"):
            # skip test for internal setting
            object.__setattr__(self, name, value)
            return

        if isinstance(value, ConfigSection):
            # child ConfigSection
            self._internal_children.add(name)
            self._internal_settings.add(name)
            value._internal_parent_ref = self
            object.__setattr__(self, name, value)
            return

        # FIXME: This needs to be fixed when we run with py3 env
        # This is the best solution we can afford right now. We tried to use
        # Utils.Utilities.encodeUnicodeToBytes but it misteriously caused
        # https://github.com/dmwm/WMCore/issues/10381
        if not PY3:
            if isinstance(value, unicode):
                value = str(value)

        # for backward compatibility use getattr and sure to work if the
        # _internal_skipChecks flag is not set
        if not getattr(self, '_internal_skipChecks', False):
            self._complexTypeCheck(name, value)

        object.__setattr__(self, name, value)
        self._internal_settings.add(name)
        return

    def __delattr__(self, name):
        if name.startswith("_internal_"):
            # skip test for internal setting
            object.__delattr__(self, name)
            return
        else:
            if name in self._internal_children:
                self._internal_children.remove(name)
            if name in self._internal_settings:
                self._internal_settings.remove(name)
            object.__delattr__(self, name)
            return

    def __iter__(self):
        for attr in self._internal_settings:
            yield getattr(self, attr)

    def __add__(self, otherSection):
        """
        _addition operator_

        Define addition for two config section objects

        """
        for setting in otherSection._internal_settings:
            settingInstance = getattr(otherSection, setting)
            if setting in self._internal_settings:
                currentSetting = getattr(self, setting)
                if not isinstance(currentSetting, type(settingInstance)) \
                        and currentSetting is not None and settingInstance is not None:
                    msg = "Trying to overwrite a setting with mismatched types"
                    msg += "%s.%s is not the same type as %s.%s" % (
                        self._internal_name, setting,
                        otherSection._internal_name, setting
                    )

                    raise TypeError(msg)
            self.__setattr__(setting, settingInstance)
        return self

    def section_(self, sectionName):
        """
        _section_

        Get a section by name, create it if not present,
        returns a ConfigSection instance

        """
        if sectionName in self.__dict__:
            return self.__dict__[sectionName]
        newSection = ConfigSection(sectionName)
        self.__setattr__(sectionName, newSection)
        return object.__getattribute__(self, sectionName)

    def pythonise_(self, **options):
        """
        convert self into list of python format strings

        options available

        document - if True will add document_ calls to the python
        comment  - if True will add docs as comments.

        """
        document = options.get('document', False)
        comment = options.get('comment', False)
        prefix = options.get('prefix', None)

        if prefix is not None:
            myName = "%s.%s" % (prefix, self._internal_name)
        else:
            myName = self._internal_name

        result = []
        if document:
            result.append("%s.document_(\"\"\"%s\"\"\")" % (
                myName,
                self._internal_documentation)
                          )
        if comment:
            result.append("# %s: %s" % (
                myName, self._internal_documentation.replace(
                        "\n", "\n# "),
            ))
        for attr in self._internal_settings:
            if attr in self._internal_children:
                result.append("%s.section_(\'%s\')" % (myName, attr))
                result.extend(getattr(self, attr).pythonise_(
                        document=document, comment=comment, prefix=myName))
                continue
            if attr in self._internal_docstrings:
                if comment:
                    result.append("# %s.%s: %s" % (
                        myName, attr,
                        self._internal_docstrings[attr].replace("\n", "\n# ")
                    ))
            result.append("%s.%s = %s" % (
                myName,
                attr, formatAsString(getattr(self, attr))
            ))

            if attr in self._internal_docstrings:
                if document:
                    result.append(
                            "%s.document_(\"\"\"%s\"\"\", \'%s\')" % (
                                myName,
                                self._internal_docstrings[attr], attr))
        return result

    def dictionary_(self):
        """
        _dictionary_

        Create a dictionary representation of this object.

        This method does not take into account possible ConfigSections
        as attributes of self (i.e. sub-ConfigSections) as the
        dictionary_whole_tree_() method does.
        The reason for this method to stay is that WebTools.Root.py
        depends on a few places to check itself like:
        if isinstance(param_value, ConfigSection) ...

        """
        result = {}
        for x in self._internal_settings:
            result.__setitem__(x, getattr(self, x))
        return result

    def dictionary_whole_tree_(self):
        """
        Create a dictionary representation of this object.

        ConfigSection.dictionary_() method needs to expand possible
        items that are ConfigSection instances (those which appear
        in the _internal_children set).
        Also these sub-ConfigSections have to be made dictionaries
        rather than putting e.g.
        'Task1': <WMCore.Configuration.ConfigSection object at 0x104ccb50>

        """
        result = {}
        for x in self._internal_settings:
            if x in self._internal_children:
                v = getattr(self, x)
                result[x] = v.dictionary_whole_tree_()
                continue
            result.__setitem__(x, getattr(self, x))  # the same as result[x] = value
        return result

    def document_(self, docstring, parameter=None):
        """
        _document_

        Add docs/comments to parameters. If the parameter is None, then
        the documentation is provided to the section itself.
        This method will overwrite any existing documentation

        """
        if parameter is None:
            self._internal_documentation = str(docstring)
            return
        self._internal_docstrings[parameter] = str(docstring)
        return

    def __str__(self):
        """
        string representation, dump to python format
        """
        result = ""
        for pystring in self.pythonise_():
            result += "%s\n" % pystring
        return result

    def documentedString_(self):
        """
        string representation, dump to python format
        include docs as calls to document_
        """
        result = ""
        for pystring in self.pythonise_(document=True):
            result += "%s\n" % pystring
        return result

    def commentedString_(self):
        """
        string representation, dump to python format
        include docs as comments
        """
        result = ""
        for pystring in self.pythonise_(comment=True):
            result += "%s\n" % pystring
        return result

    # Added by mnorman to make our strategy to use configSections viable
    def listSections_(self):
        """
        _listSections_

        Retrieve a list of components from the components
        configuration section

        """
        comps = self._internal_settings
        return list(comps)

    def getInternalName(self):
        """
        _getInternalName_

        Return the internal name

        """
        return self._internal_name


class Configuration(object):
    # pylint: disable=protected-access
    """
    _Configuration_

    Top level configuration object

    """

    def __init__(self):
        object.__init__(self)
        self._internal_components = []
        self._internal_webapps = []
        self._internal_sections = []
        Configuration._instance = self

    def __add__(self, otherConfig):
        """
        _addition operator_

        Define addition for two config section objects

        """
        for configSect in otherConfig._internal_sections:
            if configSect not in self._internal_sections:
                self.section_(configSect)
            getattr(self, configSect) + getattr(otherConfig, configSect)
        return self

    def __setattr__(self, name, value):
        if name.startswith("_internal_"):
            # skip test for internal settings
            object.__setattr__(self, name, value)
            return
        if not isinstance(value, ConfigSection):
            msg = "Can only add objects of type ConfigSection to Configuration"
            raise RuntimeError(msg)

        object.__setattr__(self, name, value)
        return

    def __delattr__(self, name):
        if name.startswith("_internal_"):
            # skip test for internal setting
            object.__delattr__(self, name)
            return
        else:
            if name in self._internal_sections:
                self._internal_sections.remove(name)
            if name in self._internal_components:
                self._internal_components.remove(name)
            if name in self._internal_webapps:
                self._internal_webapps.remove(name)
            object.__delattr__(self, name)
            return

    @staticmethod
    def getInstance():
        return getattr(Configuration, "_instance", None)

    def listComponents_(self):
        """
        _listComponents_

        Retrieve a list of components from the components
        configuration section

        """
        comps = self._internal_components
        return comps

    def listWebapps_(self):
        """
        _listWebapps_

        Retrieve a list of webapps from the webapps configuration section.
        """
        return self._internal_webapps

    def listSections_(self):
        """
        _listSections_

        Retrieve a list of components from the components
        configuration section

        """
        comps = self._internal_sections
        return comps

    def section_(self, sectionName):
        """
        _section_

        Get a section by name, create it if not present,
        returns a ConfigSection instance

        """
        if sectionName in self.__dict__:
            return self.__dict__[sectionName]
        newSection = ConfigSection(sectionName)
        self.__setattr__(sectionName, newSection)
        self._internal_sections.append(sectionName)
        return object.__getattribute__(self, sectionName)

    def component_(self, componentName):
        """
        _component_

        Get the config for the named component, add it
        if not present, returns a ConfigSection with
        default fields for the component added to it

        """
        compSection = self.section_(componentName)
        if componentName not in self._internal_components:
            self._internal_components.append(componentName)
            compSection.componentDir = None

        return compSection

    def webapp_(self, webappName):
        """
        _webapp_

        Get the config for the named webapp, add it if not present.  This will
        return a ConfigSection with default fields for the webapp added to it.
        """
        webappSection = self.section_(webappName)
        if webappName not in self._internal_webapps:
            self._internal_webapps.append(webappName)
            webappSection.section_("Webtools")
            webappSection.section_("database")
            webappSection.section_("security")

        return webappSection

    def pythonise_(self, **options):
        """
        write as python format


        document - if True will add document_ calls to the python
        comment  - if True will add docs as comments.

        """
        document = options.get('document', False)
        comment = options.get('comment', False)

        result = "from WMCore.Configuration import Configuration\n"
        result += "config = Configuration()\n"
        for sectionName in self._internal_sections:
            if sectionName in self._internal_components:
                result += "config.component_(\'%s\')\n" % sectionName
            elif sectionName in self._internal_webapps:
                result += "config.webapp_(\'%s\')\n" % sectionName
            else:
                result += "config.section_(\'%s\')\n" % sectionName

            sectionRef = getattr(self, sectionName)
            for sectionAttr in sectionRef.pythonise_(
                    document=document, comment=comment):
                if sectionAttr.startswith("#"):
                    result += "%s\n" % sectionAttr
                else:
                    result += "config.%s\n" % sectionAttr

        return result

    def __str__(self):
        """
        string format of this object

        """
        return self.pythonise_()

    def documentedString_(self):
        """
        python format with document_ calls
        """
        return self.pythonise_(document=True)

    def commentedString_(self):
        """
        python format with docs as comments
        """
        return self.pythonise_(comment=True)


class ConfigurationEx(Configuration):
    """
    _Configuration_

    Top level extended configuration object

    Allows to freely set parameters of the configuration. Things like callables
    can be used as configuration parameters now. Drawback is that the configuration
    cannot be saved and passed around using the code.
    """

    def __init__(self):
        # super(ConfigurationEx, self).__init__()
        Configuration.__init__(self)

    def section_(self, sectionName):
        """
        _section_

        Get a section by name, create it if not present,
        and set the skipChecks flag
        returns a ConfigSection instance

        """
        section = Configuration.section_(self, sectionName)
        section._internal_skipChecks = True  # pylint: disable=protected-access
        return section


def loadConfigurationFile(filename):
    """
    _loadConfigurationFile_

    Load a Configuration File

    """

    cfgBaseName = os.path.basename(filename).replace(".py", "")
    cfgDirName = os.path.dirname(filename)
    if not cfgDirName:
        modSpecs = importlib.machinery.PathFinder().find_spec(cfgBaseName)
    else:
        modSpecs = importlib.machinery.PathFinder().find_spec(cfgBaseName, [cfgDirName])
    try:
        modRef = modSpecs.loader.load_module()
    except Exception as ex:
        msg = "Unable to load Configuration File:\n"
        msg += "%s\n" % filename
        msg += "Due to error:\n"
        msg += str(ex)
        msg += str(traceback.format_exc())
        raise RuntimeError(msg)

    for attr in listvalues(modRef.__dict__):
        if isinstance(attr, Configuration):
            return attr

    # //
    # //  couldnt find a Configuration instance
    # //
    msg = "Unable to find a Configuration object instance in file:\n"
    msg += "%s\n" % filename
    raise RuntimeError(msg)


def saveConfigurationFile(configInstance, filename, **options):
    """
    _saveConfigurationFile_

    Save the configuration as a python module
    Options controls the format of documentation

    comment = True means save docs as comments
    document = True means save docs as document_ calls


    """
    if isinstance(configInstance, ConfigurationEx):
        raise NotImplementedError("ConfigurationEx instances cannot be saved. Use Configuration instead.")
    comment = options.get("comment", False)
    document = options.get("document", False)
    if document:
        comment = False

    with open(filename, 'w') as handle:
        if document:
            handle.write(configInstance.documentedString_())
        elif comment:
            handle.write(configInstance.commentedString_())
        else:
            handle.write(str(configInstance))

    return



#!/usr/bin/env python
"""
_SetStateTime_

MySQL implementation of Jobs.SetStateTime
"""




from WMCore.Database.DBFormatter import DBFormatter

class SetStateTime(DBFormatter):
    """
    _SetStateTime_

    Update the state transition time for the given job.
    """
    sql = "UPDATE wmbs_job SET state_time = :statetime WHERE id = :jobid"

    def execute(self, jobID = None, stateTime = None, conn = None,
                transaction = False):
        binds = {"jobid": jobID, "statetime": stateTime}
        self.dbi.processData(self.sql, binds,
                             conn = conn, transaction = transaction)
        return



#!/usr/bin/env python
"""
_GetLocation_

MySQL implementation of Jobs.Location
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class GetLocation(DBFormatter):
    """
    _GetLocation_

    Retrieve all files that are associated with the given job from the
    database.
    """
    sql = "SELECT site_name FROM wmbs_location WHERE id = (SELECT location FROM wmbs_job WHERE id = :jobid)"
    bulkSQL = """SELECT wmbs_location.site_name as site_name, wmbs_job.id as id
                 FROM wmbs_location
                 INNER JOIN wmbs_job
                 ON wmbs_location.id = wmbs_job.location
                 WHERE wmbs_job.id = :jobid"""

    #def format(self, results):
    #    """
    #    _formatDict_
    #
    #    """
    #
    #    if len(results) == 0:
    #        return False
    #    else:
    #        return results[0][0]


    def execute(self, jobid, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """

        if isinstance(jobid, list):
            result = self.dbi.processData(self.bulkSQL, jobid, conn = conn, transaction = transaction)
            tmp = self.formatDict(result)
            return tmp
        result = self.dbi.processData(self.sql, {"jobid": jobid}, conn = conn,
                                      transaction = transaction)

        return self.format(result)



#!/usr/bin/env python
"""
_Save_

MySQL implementation of Masks.Save
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter



class Save(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_job_mask (job, firstevent, lastevent, firstrun,
               lastrun, firstlumi, lastlumi, inclusivemask)
             VALUES (:jobid, :firstevent, :lastevent, :firstrun, :lastrun, :firstlumi,
               :lastlumi, :inclusivemask)
    """

    def execute(self, jobid, mask, conn = None, transaction = False):
        if isinstance(mask, list):
            # Bulk commit
            # Hope you didn't send us a list of empty masks
            binds = []
            for m in mask:
                binds.append({"jobid": m['jobID'], 'firstevent': m['FirstEvent'], 'lastevent': m['LastEvent'],
                              'firstrun': m['FirstRun'], 'lastrun': m['LastRun'], 'firstlumi': m['FirstLumi'],
                              'lastlumi': m['LastLumi'], 'inclusivemask': m['inclusivemask']})
        else:
            # Simple one-part mask
            binds = {"jobid": jobid, 'firstevent': mask['FirstEvent'], 'lastevent': mask['LastEvent'],
                     'firstrun': mask['FirstRun'], 'lastrun': mask['LastRun'], 'firstlumi': mask['FirstLumi'],
                     'lastlumi': mask['LastLumi'], 'inclusivemask': mask['inclusivemask']}

            fail = True
            for key in binds:
                if key != 'jobid' and key != 'inclusivemask' and binds[key] != None:
                    # At least one of the keys contains something
                    fail = False
            if fail:
                return

        # Actually run the code
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_GetAllJobs_

Oracle implementation of Jobs.GetAllJobs
"""

from WMCore.WMBS.MySQL.Jobs.GetAllJobs import GetAllJobs as MySQLGetAllJobs


class GetAllJobs(MySQLGetAllJobs):
    """
    Besides the row limitation, it's the same as in MySQL
    """
    limit_sql = " AND ROWNUM <= %d"



"""
_CheckComponentStatus_

MySQL implementation of CheckComponentStatus
"""

__all__ = []



import time
from WMCore.Database.DBFormatter import DBFormatter

class CheckComponentStatus(DBFormatter):

    sql = """SELECT comp.name as name
             FROM wm_workers worker
             INNER JOIN wm_components comp ON comp.id = worker.component_id
             INNER JOIN (SELECT component_id, MAX(last_updated) AS last_updated FROM wm_workers
                         GROUP BY component_id) max_result
                        ON (worker.last_updated = max_result.last_updated
                           AND max_result.component_id = comp.id)
             WHERE max_result.last_updated + comp.update_threshold < :current_time
             """
    #sql = """select max(last_updated) from wm_workers"""

    def execute(self,detail = False, conn = None, transaction = False):
        binds = {'current_time' : int(time.time())}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                  transaction = transaction)
        # If detail flag is true return list of components which were down
        resultList = self.format(result)
        if detail:
            componentList = []
            for a in resultList:
                componentList.append(a[0])
            return componentList

        if len(resultList) > 0:
            return {'status': 'down'}
        else:
            return {'status': 'ok'}



#!/usr/bin/env python
"""
_CompleteJob_

MySQL implementation for labeling a job Complete
"""

import logging

from WMCore.Database.DBFormatter import DBFormatter

class CompleteJob(DBFormatter):
    """
    _CompleteJob_

    Label jobs as complete
    """


    sql = """UPDATE bl_runjob SET status = '0' WHERE id = :id"""



    def execute(self, jobs, conn = None, transaction = False):
        """
        _execute_

        Complete jobs
        """

        if len(jobs) < 1:
            # Then we have nothing to do
            return

        binds = []
        for job in jobs:
            binds.append({'id': job})


        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)


        return



#!/usr/bin/env python
"""
_SiblingSubscriptionsComplete_

MySQL implementation of Subscription.SiblingSubscriptionsComplete
"""

from WMCore.Database.DBFormatter import DBFormatter


class SiblingSubscriptionsComplete(DBFormatter):
    """
    For each file in the input fileset count the number of subscriptions
    (on the same input fileset) that have completed the file. If the number
    of subscriptions that have completed the file is the same as the number
    of subscriptions that processed the file (not counting this subscription)
    we can say that processing of the file is complete and we can perform
    some other action on it (usually deletion).

    """
    sql = """SELECT wmbs_file_details.id,
                    wmbs_file_details.events,
                    wmbs_file_details.lfn,
                    wmbs_pnns.pnn
             FROM (
               SELECT wmbs_sub_files_available.fileid
               FROM wmbs_sub_files_available
                 INNER JOIN wmbs_subscription ON
                   wmbs_subscription.id = wmbs_sub_files_available.subscription
                 LEFT OUTER JOIN wmbs_subscription sibling_subscription ON
                   sibling_subscription.fileset = wmbs_subscription.fileset AND
                   sibling_subscription.id != wmbs_subscription.id
                 LEFT OUTER JOIN wmbs_sub_files_complete ON
                   wmbs_sub_files_complete.fileid = wmbs_sub_files_available.fileid AND
                   wmbs_sub_files_complete.subscription = sibling_subscription.id
               WHERE wmbs_sub_files_available.subscription = :subscription
               GROUP BY wmbs_sub_files_available.fileid
               HAVING COUNT(sibling_subscription.id) = COUNT(wmbs_sub_files_complete.fileid)
             ) available_files
             INNER JOIN wmbs_file_details ON
               wmbs_file_details.id = available_files.fileid
             INNER JOIN wmbs_file_location ON
               wmbs_file_location.fileid = available_files.fileid
             INNER JOIN wmbs_pnns ON
                 wmbs_file_location.pnn = wmbs_pnns.id
             """

    def execute(self, subscription, conn=None, transaction=False):
        results = self.dbi.processData(self.sql, {'subscription': subscription},
                                       conn=conn, transaction=transaction)

        return self.formatDict(results)



#!/usr/bin/env python
"""
_GetSiteSE_

MySQL implementation of Locations.GetSiteSE
"""

__all__ = []

from WMCore.Database.DBFormatter import DBFormatter


class GetSiteSE(DBFormatter):
    """
    Grab all the relevant information for a given site.
    Usually useful only in the submitter
    """

    sql = """SELECT wpnn.pnn FROM wmbs_location_pnns wlpnn
               INNER JOIN wmbs_location wl ON wlpnn.location = wl.id
               INNER JOIN wmbs_pnns wpnn ON wlpnn.pnn = wpnn.id
             WHERE wl.ce_name = :site"""

    def execute(self, cesite=None, conn=None, transaction=False):
        results = self.dbi.processData(self.sql, {'site': cesite}, conn=conn, transaction=transaction)
        return self.formatDict(results)



#!/usr/bin/env python
"""
_GeneratorInterface_

API definition for Generator objects

"""

from builtins import object


class GeneratorInterface(object):
    """
    _GeneratorInterface_

    Define the APIs to be implemented for Generator objects invoked
    by the JobFactory

    """
    def __init__(self, **options):
        self.task = options.get("task", None)
        self.options = options


    def start(self):
        """
        _start_

        Invoked by the factory before calling on any jobs.
        Override to initialise any state required

        """
        #print "%s.start()" % self.__class__.__name__
        pass

    def finish(self):
        """
        _finish_

        Invoked by the factory after it has completed generating
        a set of jobs.
        Override to save state as required

        """
        #print "%s.finish()" % self.__class__.__name__
        pass

    def startGroup(self, jobGroup):
        """
        _startGroup_

        Hook to respond to the start of a new job group

        """
        #print "%s.startGroup(%s)" % (self.__class__.__name__, jobGroup)
        pass

    def finishGroup(self, jobGroup):
        """
        _finishGroup_

        Hook to respond to end of a job group

        """
        #print "%s.finishGroup(%s)" % (self.__class__.__name__, jobGroup)
        pass

    def __call__(self, wmbsJob):
        """
        _operator(wmbsJob)_

        Act on a job instance and install the appropriate information
        into it

        """
        msg = "Object %s does not implement Generator.__call__(wmbsJob)" % (
            self.__class__.__name__,)
        raise NotImplementedError(msg)



#!/usr/bin/env python
# pylint: disable=R0902,R0201,W0703,E1102
"""
_BaseWorkerThread_

Base class for all regular worker threads managed by WorkerThreadManager.
Deriving classes should override algorithm, and optionally setup and terminate
to perform thread-specific setup and clean-up operations
"""

from builtins import object
import logging
import sys
import threading
import time
import traceback

from WMCore.Database.DBExceptionHandler import db_exception_handler
from WMCore.Database.Transaction import Transaction


class BaseWorkerThread(object):
    """
    A base class for worker threads, used for work that needs to occur at
    regular intervals. Framework (through WorkerThreadManager) ensures that
    a default transaction, trigger and message service are available as in
    event handler threads.
    """

    def __init__(self):
        """
        Creates the worker, called from parent thread
        """
        self.idleTime = None
        self.notifyTerminate = None
        self.notifyPause = None
        self.notifyResume = None

        # Reference to the owner component and arguments
        # this will be set when addWorker is called in WorkerThreadManager
        self.component = None
        self.args = {}
        self.heartbeatAPI = None

        # Termination callback function
        self.terminateCallback = None

        # Init heartbeat flag and worker name
        self.useHeartbeat = False
        self.workerName = None

        # Init the timing
        self.lastTime = time.time()

        # Get the current DBFactory
        myThread = threading.currentThread()
        self.dbFactory = myThread.dbFactory

        # Get the logger
        self.logger = myThread.logger

        # get the procid from the mainthread msg service
        # if we use this in testing it might not be there.
        self.procid = 0
        if hasattr(myThread, 'msgService'):
            self.procid = myThread.msgService.procid

    def setup(self, parameters):
        """
        Called when thread is being run for the first time. Optional in derived
        classes.
        """
        pass

    def terminate(self, parameters):
        """
        Called when thread is being terminated. Optional in derived classes.
        If inherited, then derived class has to call this method.
        """
        msg = "Thread gracefully terminated at %s" % time.strftime("%Y-%m-%dT%H:%M:%S")
        logging.info(msg)
        if self.useHeartbeat:
            self.heartbeatAPI.updateWorkerError(self.workerName, msg)

    def algorithm(self, parameters):
        """
        The method that performs the required work. Should be overridden in
        derived classes.

        If this method raises an exception all workers will be terminated
        """
        logging.error("Calling algorithm on BaseWorkerThread: Override me!")

    def setUpHeartbeat(self, myThread):
        # heartbeat needed to be called in self.initInThread
        # to get the right name but before the self.setup
        if hasattr(self.component.config, "Agent"):
            self.useHeartbeat = getattr(self.component.config.Agent, "useHeartbeat", True)
            self.workerName = myThread.getName()

        if self.useHeartbeat:
            self.heartbeatAPI.registerWorker(self.workerName)
        return

    def setUpLogDB(self, myThread):
        # setup logDB
        if hasattr(self.component.config, "General") and \
                hasattr(self.component.config.General, "central_logdb_url") and \
                hasattr(self.component.config, "Agent"):
            from WMCore.Services.LogDB.LogDB import LogDB
            myThread.logdbClient = LogDB(self.component.config.General.central_logdb_url,
                                         self.component.config.Agent.hostName, logger=logging)
        else:
            myThread.logdbClient = None
        return

    def initInThread(self, parameters):
        """
        Called when the thread is actually running in its own thread. Performs
        internal object setup.
        """
        # Get the DB Factory we were passed by parent thread and assign to this
        # thread
        myThread = threading.currentThread()
        myThread.name = self.__class__.__name__
        myThread.dbFactory = self.dbFactory

        # Now we're in our own thread, set the logger
        myThread.logger = self.logger

        (connectDialect, _junk) = self.component.config.CoreDatabase.connectUrl.split(":", 1)

        if connectDialect.lower() == "mysql":
            myThread.dialect = "MySQL"
        elif connectDialect.lower() == "oracle":
            myThread.dialect = "Oracle"

        logging.info("Initialising default database")
        myThread.dbi = myThread.dbFactory.connect()
        logging.info("Initialising default transaction")
        myThread.transaction = Transaction(myThread.dbi)

        self.setUpHeartbeat(myThread)
        self.setUpLogDB(myThread)

        # Call worker setup
        self.setup(parameters)
        myThread.transaction.commit()

    def __call__(self, parameters):
        """
        Thread entry point; handles synchronisation with run and terminate
        conditions
        """
        errMsg = ""
        try:
            msg = "Initialising worker thread %s" % str(self)
            logging.info(msg)

            # Call thread startup method
            self.initInThread(parameters)

            msg = "Worker thread %s started" % str(self)
            logging.info(msg)

            myThread = threading.currentThread()

            # Run event loop while termination is not flagged
            algorithmWithDBExceptionHandler = db_exception_handler(self.algorithm)
            while not self.notifyTerminate.isSet():
                # Check manager hasn't paused threads
                if self.notifyPause.isSet():
                    self.notifyResume.wait()
                else:
                    # Catch case where threads were paused and then terminated
                    #  - threads should not run in this case!
                    if not self.notifyTerminate.isSet():
                        # Do some work!
                        try:
                            if self.useHeartbeat:
                                self.heartbeatAPI.updateWorkerHeartbeat(self.workerName, "Running")

                            tSpent, results, _ = algorithmWithDBExceptionHandler(parameters)
                            if tSpent and self.useHeartbeat:
                                logging.info("%s took %.3f secs to execute", self.workerName, tSpent)
                                self.heartbeatAPI.updateWorkerCycle(self.workerName, tSpent, results)

                            # Catch if someone forgets to commit/rollback
                            if myThread.transaction.transaction is not None:
                                msg = """ Thread %s:  Transaction reached
                                          end of poll loop.""" % self.workerName
                                msg += " Raise a bug against me. Rollback."
                                logging.error(msg)
                                myThread.transaction.rollback()
                        except Exception as ex:
                            if myThread.transaction.transaction is not None:
                                myThread.transaction.rollback()
                            errMsg = "Error in worker algorithm (1):\nBacktrace:\n "
                            errMsg += (" %s %s" % (str(self), str(ex)))
                            stackTrace = traceback.format_tb(sys.exc_info()[2], None)
                            for stackFrame in stackTrace:
                                errMsg += stackFrame
                            logging.error(errMsg)
                            # force entire component to terminate
                            try:
                                self.component.prepareToStop()
                            except Exception as ex1:
                                logging.error("Failed to halt component after worker crash: %s", str(ex1))
                            raise ex
                        # Put the thread to sleep
                        self.sleepThread()

            # Call specific thread termination method
            self.terminate(parameters)
        except Exception as ex:
            # Notify error
            msg = "Error in event loop (2): %s %s\nBacktrace:\n"
            msg = msg % (str(self), str(ex))
            stackTrace = traceback.format_tb(sys.exc_info()[2], None)
            for stackFrame in stackTrace:
                msg += stackFrame
            logging.error(msg)
            # send heartbeat message that thread is terminated
            if self.useHeartbeat:
                msg = errMsg or msg
                try:
                    self.heartbeatAPI.updateWorkerError(self.workerName, errMsg)
                except Exception as ex:
                    logging.error("Heartbeat error update failed %s", str(ex))

        # Indicate to manager that thread is done
        self.terminateCallback(threading.currentThread().name)

        # All done
        msg = "Worker thread %s terminated" % str(self)
        logging.info(msg)

    def sleepThread(self):
        """
        _sleepThread_

        A subclassable method to make the thread sleep.

        The default (naiive) time.sleep(self.idleTime) isn't always
        the best idea, let different workers do it differently.

        Wakes the thread up every minute for a quick heartbeat.
        Need to constantly watch if the thread is terminated for
        properly stopping/terminating it.

        returns control when it's time to wake back up
        doesn't return any values
        """
        idleTime = self.idleTime
        while idleTime > 0:
            if self.useHeartbeat and idleTime % 60 == 0:
                # send a heartbeat every minute
                self.heartbeatAPI.updateWorkerHeartbeat(self.workerName, "Running")

            if self.notifyTerminate.isSet():
                break

            time.sleep(1)
            idleTime -= 1



#!/usr/bin/env python
# pylint: disable=W0613

"""
_RetryManager_

The retry manager picks up jobs from their cooloff state and using a set of plugins
attempts to put them in their non-cooloff state again.

"""

import logging
import threading

from WMCore.Agent.Harness import Harness
from WMComponent.RetryManager.RetryManagerPoller import RetryManagerPoller


class RetryManager(Harness):
    """
    _RetryManager_

    The retry manager picks up jobs from their cooloff state and using a set
    of plugins attempts to put them in their non-cooloff state again
    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.config = config

    def preInitialization(self):
        """
        Initializes plugins for different messages
        """

        # Add event loop to worker manager
        myThread = threading.currentThread()
        pollInterval = self.config.RetryManager.pollInterval
        logging.info("Setting poll interval to %s seconds", pollInterval)
        myThread.workerThreadManager.addWorker(RetryManagerPoller(self.config), pollInterval)



#!/usr/bin/env python
"""
_ExistsByID_

Oracle implementation of JobGroup.ExistsByID
"""

__all__ = []



from WMCore.WMBS.MySQL.JobGroup.ExistsByID import ExistsByID as ExistsByIDJobGroupMySQL

class ExistsByID(ExistsByIDJobGroupMySQL):
    sql = ExistsByIDJobGroupMySQL.sql



#!/usr/bin/env python
"""
_FailedJobs_

MySQL implementation of Workflow.FailedJobs
"""




from WMCore.Database.DBFormatter import DBFormatter

class FailedJobs(DBFormatter):
    sql = """SELECT wmbs_users.cert_dn as owner, wmbs_workflow.task, wmbs_job.id
                    FROM wmbs_workflow
               INNER JOIN wmbs_users ON
                 wmbs_workflow.owner = wmbs_users.id
               INNER JOIN wmbs_subscription ON
                 wmbs_workflow.id = wmbs_subscription.workflow
               LEFT OUTER JOIN wmbs_jobgroup ON
                 wmbs_subscription.id = wmbs_jobgroup.subscription
               LEFT OUTER JOIN wmbs_job ON
                 wmbs_jobgroup.id = wmbs_job.jobgroup
               LEFT OUTER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
             WHERE wmbs_job.outcome = 0 AND
                   (wmbs_job_state.name = 'exhausted' OR
                    wmbs_job_state.name = 'cleanout')"""

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_GlobalWorkQueueConfig_

Global WorkQueue config.
WARNING: This config is only used for wmagent all-in-one test. Will be deplicated.
Doesn't replex the current workqueue config which is under deployment
"""

import os

from WMCore.Configuration import Configuration

# The following parameters may need to be changed but nothing else in the config
# will.  The server host name needs to match the machine this is running on, the
# port numbers will not need to be changed unless there is more than one
# ReqMgr/GlobalWorkQueue running on the machine.
serverHostName = "SERVER_HOSTNAME"
reqMgrHostName = "REQMGR_HOSTNAME"
reqMgrPort = 8687

# The work directory and database need to be separate from the WMAgent
# installation.
workDirectory = "WORKQUEUE_WORK_DIRECTORY"
databaseUrl = "mysql://DBSUSER:DBSPASSWORD@localhost/ReqMgrDB"
databaseSocket = "/opt/MySQL-5.1/var/lib/mysql/mysql.sock"

# The couch username and password needs to be added.  The GroupUser and
# ConfigCache couch apps need to be installed into the configcache couch
# database.  The JobDump couchapp needs to be installed into the jobdump
# database.  The GroupUser and ACDC couchapps needs to be install into the
# acdc database.
couchURL = "http://USERNAME:PASSWORD@COUCHSERVER:5984"
workqueueDBName = 'workqueue'
workqueueInboxDbName = 'workqueue_inbox'
wmstatDBName = "wmstats"

# Agent name and team name.
agentName = "WMAgent"
teamName = "cmsdataops"
contactName = "cmsdataops@cern.ch"

# Nothing after this point should need to be changed.
config = Configuration()

config.section_("Agent")
config.Agent.hostName = serverHostName
config.Agent.teamName = teamName
config.Agent.agentName = agentName
config.Agent.useMsgService = False
config.Agent.useTrigger = False
config.Agent.useHeartbeat = False
config.Agent.contact = contactName

config.section_("General")
config.General.workDir = workDirectory

config.section_("CoreDatabase")
config.CoreDatabase.connectUrl = databaseUrl
config.CoreDatabase.socket = databaseSocket

reqMgrUrl = "http://%s:%s" % (reqMgrHostName, reqMgrPort)

config.component_("WorkQueueManager")
config.WorkQueueManager.namespace = "WMComponent.WorkQueueManager.WorkQueueManager"
config.WorkQueueManager.componentDir = os.path.join(config.General.workDir, "WorkQueueManager")
config.WorkQueueManager.couchurl = couchURL
config.WorkQueueManager.dbname = workqueueDBName
config.WorkQueueManager.wmstatDBName = wmstatDBName
config.WorkQueueManager.inboxDatabase = workqueueInboxDbName
config.WorkQueueManager.level = "GlobalQueue"
config.WorkQueueManager.pollInterval = 600
config.WorkQueueManager.queueParams = {'WMStatsCouchUrl': "%s/%s" % (config.WorkQueueManager.couchurl.rstrip(), config.WorkQueueManager.wmstatDBName)}
config.WorkQueueManager.reqMgrConfig = {'teamName' : config.Agent.teamName,
                                        'endpoint': "%s/reqMgr/" % reqMgrUrl}



#!/usr/bin/env python
# Turn off to many arguments
# pylint: disable=R0913
"""
_Fileset_

A simple object representing a Fileset in WMBS.

A fileset is a collection of files for processing. This could be a
complete block, a block in transfer, some user defined dataset etc.

workflow + fileset = subscription
"""

import logging

from WMCore.DataStructs.Fileset import Fileset as WMFileset
from WMCore.WMBS.File import File, addFilesToWMBSInBulk
from WMCore.WMBS.WMBSBase import WMBSBase


class Fileset(WMBSBase, WMFileset):
    """
    A simple object representing a Fileset in WMBS.

    A fileset is a collection of files for processing. This could be a
    complete block, a block in transfer, some user defined dataset, a
    many file lumi-section etc.

    workflow + fileset = subscription
    """

    def __init__(self, name=None, id=-1, is_open=True, files=None,
                 parents=None, parents_open=True, source=None, sourceUrl=None):
        WMBSBase.__init__(self)
        WMFileset.__init__(self, name=name, files=files)

        if parents is None:
            parents = set()

        # Create a new fileset
        self.id = id
        self.open = is_open
        self.parents = parents
        self.setParentage(parents, parents_open)
        self.source = source
        self.sourceUrl = sourceUrl
        self.lastUpdate = 0

    def setLastUpdate(self, timeUpdate):
        """
        _setLastUpdate_

        Change the last update time of this fileset.  The lastUpdate parameter is a int
        representing the last time where the fileset was modifed.
        """
        closeAction = self.daofactory(classname="Fileset.SetLastUpdate")
        closeAction.execute(fileset=self.name, timeUpdate=timeUpdate, conn=self.getDBConn(),
                            transaction=self.existingTransaction())

        self.lastUpdate = timeUpdate
        return

    def addFile(self, file):
        """
        Add the file object to the set, but don't commit to the database
        Call commit() to do that - enables bulk operations
        """
        WMFileset.addFile(self, file)

    def setParentage(self, parents, parents_open):
        """
        Set parentage for this fileset - set parents to closed
        """
        if parents:
            for parent in parents:
                if isinstance(parent, Fileset):
                    self.parents.add(parent)
                else:
                    self.parents.add(Fileset(name=parent,
                                             is_open=parents_open,
                                             parents_open=False))

    def exists(self):
        """
        Does a fileset exist with this name in the database
        """
        if self.id != -1:
            action = self.daofactory(classname="Fileset.ExistsByID")
            result = action.execute(id=self.id, conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
        else:
            action = self.daofactory(classname="Fileset.Exists")
            result = action.execute(self.name, conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
            if result is not False:
                self.id = result

        return result

    def create(self):
        """
        Add the new fileset to WMBS, and commit the files
        """
        if self.exists() is not False:
            self.load()
            return

        existingTransaction = self.beginTransaction()
        createAction = self.daofactory(classname="Fileset.New")
        createAction.execute(self.name, self.open, conn=self.getDBConn(),
                             transaction=self.existingTransaction())
        self.commit()
        self.loadData()
        self.commitTransaction(existingTransaction)
        logging.info("Fileset created: %s", self.name)
        return

    def delete(self):
        """
        Remove this fileset from WMBS
        """
        action = self.daofactory(classname="Fileset.Delete")
        result = action.execute(name=self.name, conn=self.getDBConn(),
                                transaction=self.existingTransaction())

        return result

    def load(self):
        """
        _load_

        Load the name, id and time that fileset was last updated in the
        database.
        """
        if self.id > 0:
            action = self.daofactory(classname="Fileset.LoadFromID")
            result = action.execute(fileset=self.id,
                                    conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
        else:
            action = self.daofactory(classname="Fileset.LoadFromName")
            result = action.execute(fileset=self.name,
                                    conn=self.getDBConn(),
                                    transaction=self.existingTransaction())

        self.id = result["id"]
        self.name = result["name"]
        self.open = result["open"]
        self.lastUpdate = result["last_update"]

        return self

    def loadData(self, parentage=1):
        """
        _loadData_

        Load all the files that belong to this fileset.
        """
        existingTransaction = self.beginTransaction()

        if self.name is None or self.id < 0:
            self.load()

        action = self.daofactory(classname="Files.InFileset")
        results = action.execute(fileset=self.id,
                                 conn=self.getDBConn(),
                                 transaction=self.existingTransaction())

        self.files = set()
        self.newfiles = set()

        for result in results:
            thisFile = File(id=result["fileid"])
            thisFile.loadData(parentage=parentage)
            self.files.add(thisFile)

        self.commitTransaction(existingTransaction)
        return

    def commit(self):
        """
        Add contents of self.newfiles to the database,
        empty self.newfiles, reload self
        """
        existingTransaction = self.beginTransaction()

        if not self.exists():
            self.create()

        ids = []
        while len(self.newfiles) > 0:
            # Check file objects exist in the database, save those that don't
            f = self.newfiles.pop()
            if not f.exists():
                f.create()
            ids.append(f["id"])
            self.files.add(f)

        # Add Files to DB only if there are any files on newfiles
        if len(ids) > 0:
            addAction = self.daofactory(classname="Files.AddToFilesetByIDs")
            addAction.execute(file=ids, fileset=self.id,
                              conn=self.getDBConn(),
                              transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return

    def markOpen(self, isOpen):
        """
        _markOpen_

        Change the open status of this fileset.  The isOpen parameter is a bool
        representing whether or not the fileset is open.
        """
        closeAction = self.daofactory(classname="Fileset.MarkOpen")
        closeAction.execute(fileset=self.name, isOpen=isOpen,
                            conn=self.getDBConn(),
                            transaction=self.existingTransaction())
        self.open = isOpen

        return

    def __str__(self):
        """
        __str__

        Write out something useful because Fileset doesn't
        inherit from dict
        """

        st = {'name': self.name, 'files': self.files, 'id': self.id,
              'open': self.open, 'parents': self.parents,
              'lastUpdate': self.lastUpdate}

        return str(st)

    def addFilesToWMBSInBulk(self, files, workflowName, isDBS=True):
        """
        _addFilesToWMBSInBulk

        Do a bulk addition of files into WMBS. This is a speedup.
        """
        # Can / should we move this to commit???
        files = addFilesToWMBSInBulk(self.id, workflowName, files,
                                     isDBS=isDBS,
                                     conn=self.getDBConn(),
                                     transaction=self.existingTransaction())
        return files
